"""Test analytical calculation of gradients of the target function versus finite
difference calculations"""


from __future__ import annotations


def test(args=[]):
    # Python and cctbx imports
    import random
    from math import pi

    from cctbx.sgtbx import space_group, space_group_symbols

    # We will set up a mock scan and a mock experiment list
    from dxtbx.model import ScanFactory
    from dxtbx.model.experiment_list import Experiment, ExperimentList
    from libtbx.phil import parse
    from libtbx.test_utils import approx_equal
    from scitbx import matrix
    from scitbx.array_family import flex

    from dials.algorithms.refinement.parameterisation.beam_parameters import (
        BeamParameterisation,
    )
    from dials.algorithms.refinement.parameterisation.crystal_parameters import (
        CrystalOrientationParameterisation,
        CrystalUnitCellParameterisation,
    )

    # Model parameterisations
    from dials.algorithms.refinement.parameterisation.detector_parameters import (
        DetectorParameterisationSinglePanel,
    )

    # Parameterisation of the prediction equation
    from dials.algorithms.refinement.parameterisation.prediction_parameters import (
        XYPhiPredictionParameterisation,
    )
    from dials.algorithms.refinement.prediction.managed_predictors import (
        ScansExperimentsPredictor,
        ScansRayPredictor,
    )
    from dials.algorithms.refinement.reflection_manager import ReflectionManager

    # Imports for the target function
    from dials.algorithms.refinement.target import (
        LeastSquaresPositionalResidualWithRmsdCutoff,
    )

    # Reflection prediction
    from dials.algorithms.spot_prediction import IndexGenerator, ray_intersection

    from . import geometry_phil

    # Experimental model builder
    from .setup_geometry import Extract

    # Local functions
    def random_direction_close_to(vector, sd=0.5):
        return vector.rotate_around_origin(
            matrix.col((random.random(), random.random(), random.random())).normalize(),
            random.gauss(0, sd),
            deg=True,
        )

    #############################
    # Setup experimental models #
    #############################

    # make a small cell to speed up calculations
    overrides = """geometry.parameters.crystal.a.length.range = 10 15
  geometry.parameters.crystal.b.length.range = 10 15
  geometry.parameters.crystal.c.length.range = 10 15"""

    master_phil = parse(geometry_phil)

    models = Extract(master_phil, overrides, cmdline_args=args)

    mydetector = models.detector
    mygonio = models.goniometer
    mycrystal = models.crystal
    mybeam = models.beam

    # Build a mock scan for a 180 degree sequence of 0.1 degree images
    sf = ScanFactory()
    myscan = sf.make_scan(
        image_range=(1, 1800),
        exposure_times=0.1,
        oscillation=(0, 0.1),
        epochs=list(range(1800)),
        deg=True,
    )
    sequence_range = myscan.get_oscillation_range(deg=False)
    im_width = myscan.get_oscillation(deg=False)[1]
    assert sequence_range == (0.0, pi)
    assert approx_equal(im_width, 0.1 * pi / 180.0)

    experiments = ExperimentList()
    experiments.append(
        Experiment(
            beam=mybeam,
            detector=mydetector,
            goniometer=mygonio,
            scan=myscan,
            crystal=mycrystal,
            imageset=None,
        )
    )

    ###########################
    # Parameterise the models #
    ###########################

    det_param = DetectorParameterisationSinglePanel(mydetector)
    s0_param = BeamParameterisation(mybeam, mygonio)
    xlo_param = CrystalOrientationParameterisation(mycrystal)
    xluc_param = CrystalUnitCellParameterisation(mycrystal)

    ########################################################################
    # Link model parameterisations together into a parameterisation of the #
    # prediction equation                                                  #
    ########################################################################

    pred_param = XYPhiPredictionParameterisation(
        experiments, [det_param], [s0_param], [xlo_param], [xluc_param]
    )

    ################################
    # Apply known parameter shifts #
    ################################

    # shift detector by 0.2 mm each translation and 2 mrad each rotation
    det_p_vals = det_param.get_param_vals()
    p_vals = [a + b for a, b in zip(det_p_vals, [2.0, 2.0, 2.0, 2.0, 2.0, 2.0])]
    det_param.set_param_vals(p_vals)

    # shift beam by 2 mrad in one axis
    s0_p_vals = s0_param.get_param_vals()
    p_vals = list(s0_p_vals)
    p_vals[1] += 2.0
    s0_param.set_param_vals(p_vals)

    # rotate crystal a bit (=2 mrad each rotation)
    xlo_p_vals = xlo_param.get_param_vals()
    p_vals = [a + b for a, b in zip(xlo_p_vals, [2.0, 2.0, 2.0])]
    xlo_param.set_param_vals(p_vals)

    #############################
    # Generate some reflections #
    #############################

    # All indices in a 2.0 Angstrom sphere
    resolution = 2.0
    index_generator = IndexGenerator(
        mycrystal.get_unit_cell(),
        space_group(space_group_symbols(1).hall()).type(),
        resolution,
    )
    indices = index_generator.to_array()

    # Predict rays within the sequence range
    ray_predictor = ScansRayPredictor(experiments, sequence_range)
    obs_refs = ray_predictor(indices)

    # Take only those rays that intersect the detector
    intersects = ray_intersection(mydetector, obs_refs)
    obs_refs = obs_refs.select(intersects)

    # Make a reflection predictor and re-predict for all these reflections. The
    # result is the same, but we gain also the flags and xyzcal.px columns
    ref_predictor = ScansExperimentsPredictor(experiments)
    obs_refs["id"] = flex.int(len(obs_refs), 0)
    obs_refs = ref_predictor(obs_refs)

    # Set 'observed' centroids from the predicted ones
    obs_refs["xyzobs.mm.value"] = obs_refs["xyzcal.mm"]

    # Invent some variances for the centroid positions of the simulated data
    im_width = 0.1 * pi / 180.0
    px_size = mydetector[0].get_pixel_size()
    var_x = flex.double(len(obs_refs), (px_size[0] / 2.0) ** 2)
    var_y = flex.double(len(obs_refs), (px_size[1] / 2.0) ** 2)
    var_phi = flex.double(len(obs_refs), (im_width / 2.0) ** 2)
    obs_refs["xyzobs.mm.variance"] = flex.vec3_double(var_x, var_y, var_phi)

    ###############################
    # Undo known parameter shifts #
    ###############################

    s0_param.set_param_vals(s0_p_vals)
    det_param.set_param_vals(det_p_vals)
    xlo_param.set_param_vals(xlo_p_vals)

    #####################################
    # Select reflections for refinement #
    #####################################

    refman = ReflectionManager(obs_refs, experiments)

    ##############################
    # Set up the target function #
    ##############################

    # Redefine the reflection predictor to use the type expected by the Target class
    ref_predictor = ScansExperimentsPredictor(experiments)

    mytarget = LeastSquaresPositionalResidualWithRmsdCutoff(
        experiments, ref_predictor, refman, pred_param, restraints_parameterisation=None
    )

    # get the functional and gradients
    mytarget.predict()
    L, dL_dp, curvs = mytarget.compute_functional_gradients_and_curvatures()

    ####################################
    # Do FD calculation for comparison #
    ####################################

    # function for calculating finite difference gradients of the target function
    def get_fd_gradients(target, pred_param, deltas):
        """Calculate centered finite difference gradients for each of the
        parameters of the target function.

        "deltas" must be a sequence of the same length as the parameter list, and
        contains the step size for the difference calculations for each parameter.
        """

        p_vals = pred_param.get_param_vals()
        assert len(deltas) == len(p_vals)
        fd_grad = []
        fd_curvs = []

        for i in range(len(deltas)):
            val = p_vals[i]

            p_vals[i] -= deltas[i] / 2.0
            pred_param.set_param_vals(p_vals)
            target.predict()

            rev_state = target.compute_functional_gradients_and_curvatures()

            p_vals[i] += deltas[i]
            pred_param.set_param_vals(p_vals)

            target.predict()

            fwd_state = target.compute_functional_gradients_and_curvatures()

            # finite difference estimation of first derivatives
            fd_grad.append((fwd_state[0] - rev_state[0]) / deltas[i])

            # finite difference estimation of curvatures, using the analytical
            # first derivatives
            fd_curvs.append((fwd_state[1][i] - rev_state[1][i]) / deltas[i])

            # set parameter back to centred value
            p_vals[i] = val

        # return to the initial state
        pred_param.set_param_vals(p_vals)

        return fd_grad, fd_curvs

    # test normalised differences between FD and analytical calculations
    fdgrads = get_fd_gradients(mytarget, pred_param, [1.0e-7] * len(pred_param))
    diffs = [a - b for a, b in zip(dL_dp, fdgrads[0])]
    norm_diffs = tuple([a / b for a, b in zip(diffs, fdgrads[0])])
    for e in norm_diffs:
        assert abs(e) < 0.001  # check differences less than 0.1%

    # test normalised differences between FD curvatures and analytical least
    # squares approximation. We don't expect this to be especially close
    if curvs:
        diffs = [a - b for a, b in zip(curvs, fdgrads[1])]
        norm_diffs = tuple([a / b for a, b in zip(diffs, fdgrads[1])])
        for e in norm_diffs:
            assert abs(e) < 0.1  # check differences less than 10%



from __future__ import annotations

import numpy as np
import wx

from dials.array_family import flex
from dials_viewer_ext import rgb_img


class wxbmp_from_np_array:
    def __init__(
        self, lst_data_in, show_nums=True, palette="black2white", lst_data_mask_in=None
    ):
        self.wx_bmp_arr = rgb_img()
        if lst_data_in is None and lst_data_mask_in is None:
            self._ini_wx_bmp_lst = None

        else:
            self._ini_wx_bmp_lst = []
            for lst_pos in range(len(lst_data_in)):
                data_3d_in = lst_data_in[lst_pos]
                xmax = data_3d_in.shape[1]
                ymax = data_3d_in.shape[2]

                # remember to put here some assertion to check that
                # both arrays have the same shape

                if lst_data_mask_in is not None:
                    data_3d_in_mask = lst_data_mask_in[lst_pos]

                self.vl_max = float(np.amax(data_3d_in))
                self.vl_min = float(np.amin(data_3d_in))
                tmp_data2d = np.zeros((xmax, ymax), "double")
                tmp_data2d_mask = np.zeros((xmax, ymax), "double")
                z_dp = data_3d_in.shape[0]
                single_block_lst_01 = []
                for z in range(z_dp):
                    # print "z =", z
                    tmp_data2d[:, :] = data_3d_in[z : z + 1, :, :]
                    if lst_data_mask_in is not None:
                        tmp_data2d_mask[:, :] = data_3d_in_mask[z : z + 1, :, :]
                    else:
                        tmp_data2d_mask = None

                    data_sigle_img = self._wx_img_w_cpp(
                        tmp_data2d, show_nums, palette, tmp_data2d_mask
                    )

                    single_block_lst_01.append(data_sigle_img)

                self._ini_wx_bmp_lst.append(single_block_lst_01)

    def bmp_lst_scaled(self, scale=1.0):
        if self._ini_wx_bmp_lst is None:

            NewW = 350

            wx_image = wx.Image(NewW, NewW)
            wxBitmap = wx_image.ConvertToBitmap()

            dc = wx.MemoryDC(wxBitmap)
            text = "No Shoebox data"
            w, h = dc.GetSize()
            tw, th = dc.GetTextExtent(text)
            dc.Clear()
            dc.DrawText(text, (w - tw) / 2, (h - th) / 2)  # display text in center
            dc.SelectObject(wxBitmap)
            del dc
            wx_bmp_lst = [[wxBitmap]]

        else:
            wx_bmp_lst = []
            for data_3d in self._ini_wx_bmp_lst:
                single_block_lst = []
                for sigle_img_data in data_3d:
                    single_block_lst.append(self._wx_bmp_scaled(sigle_img_data, scale))

                wx_bmp_lst.append(single_block_lst)

        return wx_bmp_lst

    def _wx_img_w_cpp(self, np_2d_tmp, show_nums, palette, np_2d_mask=None):

        xmax = np_2d_tmp.shape[1]
        ymax = np_2d_tmp.shape[0]

        if np_2d_mask is None:
            np_2d_mask = np.zeros((ymax, xmax), "double")

        transposed_data = np.zeros((ymax, xmax), "double")
        transposed_mask = np.zeros((ymax, xmax), "double")

        transposed_data[:, :] = np_2d_tmp
        transposed_mask[:, :] = np_2d_mask

        flex_data_in = flex.double(transposed_data)
        flex_mask_in = flex.double(transposed_mask)

        if palette == "black2white":
            palette_num = 1
        elif palette == "white2black":
            palette_num = 2
        elif palette == "hot ascend":
            palette_num = 3
        else:  # assuming "hot descend"
            palette_num = 4

        img_array_tmp = self.wx_bmp_arr.gen_bmp(
            flex_data_in, flex_mask_in, show_nums, palette_num
        )

        np_img_array = img_array_tmp.as_numpy_array()

        height = np.size(np_img_array[:, 0:1, 0:1])
        width = np.size(np_img_array[0:1, :, 0:1])
        img_array = np.empty((height, width, 3), "uint8")
        img_array[:, :, :] = np_img_array[:, :, :]

        self._wx_image = wx.Image(width, height)
        self._wx_image.SetData(img_array.tostring())

        data_to_become_bmp = (self._wx_image, width, height)

        return data_to_become_bmp

    def _wx_bmp_scaled(self, data_to_become_bmp, scale):
        to_become_bmp = data_to_become_bmp[0]
        width = data_to_become_bmp[1]
        height = data_to_become_bmp[2]

        NewW = int(width * scale)
        NewH = int(height * scale)
        to_become_bmp = to_become_bmp.Scale(NewW, NewH, wx.IMAGE_QUALITY_NORMAL)

        wxBitmap = to_become_bmp.ConvertToBitmap()

        return wxBitmap



from __future__ import annotations

import math

import wx
import wx.grid as gridlib
import wx.lib.scrolledpanel as scroll_pan

from dials.array_family import flex
from dials.viewer.from_flex_to_wxbitmap import wxbitmap_convert


class grid_frame(wx.Frame):
    def __init__(self, parent, title):
        super().__init__(parent, title=title, size=wx.DefaultSize)

    def frame_ini_img(self, in_upper_panel, text_data=None):
        self.img_panel = in_upper_panel

        if text_data is not None:
            self.myGrid = text_data

        self.my_sizer = wx.BoxSizer(wx.VERTICAL)

        if text_data is not None:
            self.my_sizer.Add(
                self.myGrid, proportion=3, flag=wx.EXPAND | wx.ALL, border=3
            )

        self.my_sizer.SetMinSize((900, 600))
        self.SetSizer(self.my_sizer)
        self.my_sizer.Fit(self)
        self.Bind(wx.EVT_CLOSE, self.OnCloseWindow)

    def OnCloseWindow(self, event):
        wx.Exit()


class flex_3d_frame(wx.Frame):
    def __init__(self, parent, title):
        super().__init__(parent, title=title, size=wx.DefaultSize)
        self.table_exist = False

    def frame_ini_img(self, in_upper_panel, text_data=None):
        self.img_panel = in_upper_panel
        self.myGrid = None

        if text_data is not None:
            self.myGrid = text_data

        self.my_sizer = wx.BoxSizer(wx.VERTICAL)
        self.my_sizer.Add(
            self.img_panel, proportion=2, flag=wx.EXPAND | wx.ALL, border=3
        )

        if text_data is not None:
            self.my_sizer.Add(
                self.myGrid, proportion=3, flag=wx.EXPAND | wx.ALL, border=3
            )

        self.my_sizer.SetMinSize((400, 200))
        self.SetSizer(self.my_sizer)
        self.my_sizer.Fit(self)
        self.Bind(wx.EVT_CLOSE, self.OnCloseWindow)

    def OnCloseWindow(self, event):
        if not self.table_exist:
            # print "from flex_3d_frame self.myGrid = None"
            event.Skip()
        else:
            # print "from flex_3d_frame self.myGrid .NEQ. None"
            wx.Exit()


class TupTable(gridlib.GridTableBase):
    def __init__(self, data, rowLabels=None, colLabels=None):
        super().__init__()
        self.data = data
        self.rowLabels = rowLabels
        self.colLabels = colLabels

    def GetNumberRows(self):
        return len(self.data)

    def GetNumberCols(self):
        return len(self.data[0])

    def GetColLabelValue(self, col):
        if self.colLabels:
            return self.colLabels[col]

    def GetRowLabelValue(self, row):
        if self.rowLabels:
            return str(self.rowLabels[row])

    def IsEmptyCell(self, row, col):
        return False

    def GetValue(self, row, col):
        return self.data[row][col]

    def SetValue(self, row, col, value):
        pass


class MyGrid(gridlib.Grid):
    def __init__(self, parent_frame):

        self.parent_fr = parent_frame
        super().__init__(parent_frame)

    def ini_n_intro(self, table_in):
        self.lst_keys = []
        self.data = []
        self.sorted_flags = []

        for key in table_in.keys():
            if key != "shoebox":
                col_label = str(key)
                col_content = list(map(str, table_in[key]))

                str_len = bigger_size(col_label, col_content)

                if str_len > len(col_label):
                    lng_dif = int(float(str_len - len(col_label)) * 1.6)
                    add_str = " " * lng_dif
                    col_label = add_str + col_label + add_str
                else:
                    col_label = "   " + col_label + "   "

                self.lst_keys.append(col_label)
                self.data.append(col_content)
                self.sorted_flags.append(True)

        self.lst_keys.append("lst pos")
        self.data.append(list(range(len(table_in))))
        self.sorted_flags.append(True)

        self.last_col_num = len(self.lst_keys) - 1

        self.data = tuple(zip(*self.data))
        self.EnableEditing(True)
        self.EnableDragGridSize(True)

        self.set_my_table(self.last_col_num)

        self.Bind(gridlib.EVT_GRID_CELL_LEFT_CLICK, self.OnCellLeftClick)
        self.Bind(gridlib.EVT_GRID_LABEL_LEFT_CLICK, self.OnLabelLeftClick)

    def OnLabelLeftClick(self, evt):
        if evt.GetCol() == -1:
            self.repaint_img(evt.GetRow())
        else:
            self.set_my_table(evt.GetCol())

        evt.Skip()

    def set_my_table(self, col_to_sort):
        self.sorted_flags[col_to_sort] = not self.sorted_flags[col_to_sort]

        try:
            tupldata = sorted(
                self.data,
                key=lambda x: int(x[col_to_sort]),
                reverse=self.sorted_flags[col_to_sort],
            )
            print("using key=lambda x: int(x[col_to_sort])")

        except Exception:
            try:
                tupldata = sorted(
                    self.data,
                    key=lambda x: float(x[col_to_sort]),
                    reverse=self.sorted_flags[col_to_sort],
                )
                print("using key=lambda x: float(x[col_to_sort])")
            except Exception:
                tupldata = sorted(
                    self.data,
                    key=lambda x: tuple(eval(str(x[col_to_sort]))),
                    reverse=self.sorted_flags[col_to_sort],
                )
                print("using key=lambda x: tuple(x[col_to_sort])")

        colLabels = tuple(self.lst_keys)
        rowLabels = tuple(range(len(tupldata)))

        tableBase = TupTable(tupldata, rowLabels, colLabels)

        # self.AutoSizeColumns(False)
        self.SetTable(tableBase, takeOwnership=True)
        self.Refresh()
        # self.AutoSizeColumn(1)
        for i in range(len(self.lst_keys)):
            self.AutoSizeColLabelSize(i)

        # self.AutoSizeColumns(True)

    def OnCellLeftClick(self, evt):
        self.repaint_img(evt.GetRow())
        evt.Skip()

    def repaint_img(self, clikd_row):
        new_row = int(self.GetCellValue(clikd_row, self.last_col_num))
        self.parent_fr.img_panel.update_img_w_row_pos(new_row)


class flex_arr_img_panel(wx.Panel):
    def __init__(self, parent_frame):
        super().__init__(parent_frame)
        self.show_nums = True
        self.show_mask = True
        self.palette = "black2white"
        self.row_pos = 0
        self.Pframe = parent_frame
        self.local_bbox = (0, 0, 2, 2, 4, 4)

    def ini_n_intro(self, data_in_one, data_in_two=None):
        self.scale = 1.0

        if isinstance(data_in_one, flex.reflection_table):
            self.table = data_in_one
            self.assign_row_pos()

        else:
            self.first_lst_in, self.segn_lst_in = data_in_one, data_in_two
            print("flex array entered")
            self.local_bbox = None

        self.bmp_lst = self._mi_list_of_wxbitmaps()
        self.panel_left = buttons_panel(self)
        self.panel_right = multi_img_scrollable(self, self.bmp_lst)
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(self.panel_left, 0, wx.EXPAND)
        sizer.Add(self.panel_right, 1, wx.EXPAND)
        self.SetSizer(sizer)
        self.Show(True)

    def assign_row_pos(self):
        try:
            self.first_lst_in, self.segn_lst_in = (
                self.table[self.row_pos]["shoebox"].data,
                self.table[self.row_pos]["shoebox"].mask,
            )

            self.local_bbox = self.table[self.row_pos]["bbox"]

        except Exception:
            self.first_lst_in, self.segn_lst_in = None, None

    def _mi_list_of_wxbitmaps(self, re_scaling=False):
        if not re_scaling:
            if self.show_mask:
                self.lst_bmp_obj = wxbitmap_convert(self.first_lst_in, self.segn_lst_in)
            else:
                self.lst_bmp_obj = wxbitmap_convert(self.first_lst_in, None)
            return self.lst_bmp_obj.get_wxbitmap_lst(
                show_nums=self.show_nums, palette=self.palette, scale=self.scale
            )

        else:
            return self.lst_bmp_obj.scaling(scale=self.scale)

    def update_img_w_row_pos(self, num):
        self.row_pos = num
        self.assign_row_pos()
        self.bmp_lst = self._mi_list_of_wxbitmaps()
        self.panel_right.img_refresh(self.bmp_lst)

    def to_hide_nums(self):
        self.show_nums = False
        self.bmp_lst = self._mi_list_of_wxbitmaps()
        self.panel_right.img_refresh(self.bmp_lst)

    def to_show_nums(self):
        self.show_nums = True
        self.bmp_lst = self._mi_list_of_wxbitmaps()
        self.panel_right.img_refresh(self.bmp_lst)

    def to_show_mask(self):
        self.show_mask = True
        self.bmp_lst = self._mi_list_of_wxbitmaps()
        self.panel_right.img_refresh(self.bmp_lst)

    def to_hide_mask(self):
        self.show_mask = False
        self.bmp_lst = self._mi_list_of_wxbitmaps()
        self.panel_right.img_refresh(self.bmp_lst)

    def to_change_palette(self, palette_name=None):
        if palette_name is None:
            print("Something went wrong")
        else:
            self.palette = palette_name
            self.bmp_lst = self._mi_list_of_wxbitmaps()
            self.panel_right.img_refresh(self.bmp_lst)

    def to_re_zoom(self, rot_sn):
        if rot_sn > 0:
            for ntimes in range(int(math.fabs(rot_sn))):
                self.scale = self.scale * 1.05
                if self.scale > 3.0:
                    # Maximum possible zoom reached
                    self.scale = 3.0

        elif rot_sn < 0:
            for ntimes in range(int(math.fabs(rot_sn))):
                self.scale = self.scale * 0.95
                if self.scale < 0.2:
                    # Minimum possible zoom reached
                    self.scale = 0.2

        self.bmp_lst = self._mi_list_of_wxbitmaps(re_scaling=True)
        self.panel_right.img_refresh(self.bmp_lst)


class multi_img_scrollable(scroll_pan.ScrolledPanel):
    def __init__(self, outer_panel, i_bmp_in):
        super().__init__(outer_panel)
        self.parent_panel = outer_panel
        self.lst_2d_bmp = i_bmp_in
        self.SetBackgroundColour(wx.Colour(200, 200, 200))

        self.mainSizer = wx.BoxSizer(wx.VERTICAL)

        self.n_img = 0
        self.img_lst_v_sizer = wx.BoxSizer(wx.VERTICAL)

        self.set_scroll_content()

        self.mainSizer.Add(self.img_lst_v_sizer, 0, wx.CENTER | wx.ALL, 10)
        self.SetSizer(self.mainSizer)

        self.SetupScrolling()
        self.SetScrollRate(1, 1)
        self.Mouse_Pos_x = -1
        self.Mouse_Pos_y = -1
        self.old_Mouse_Pos_x = -1
        self.old_Mouse_Pos_y = -1
        self.Bind(wx.EVT_MOUSEWHEEL, self.OnMouseWheel)
        self.Bind(wx.EVT_MOTION, self.OnMouseMotion)
        self.Bind(wx.EVT_LEFT_DOWN, self.OnLeftButDown)
        self.Bind(wx.EVT_LEFT_UP, self.OnLeftButUp)
        self.Bind(wx.EVT_IDLE, self.OnIdle)
        self.scroll_rot = 0

    def OnLeftButDown(self, event):
        self.Bdwn = True
        self.old_Mouse_Pos_x, self.old_Mouse_Pos_y = event.GetPosition()

    def OnLeftButUp(self, event):
        self.Bdwn = False
        _ = event.GetPosition()

    def set_scroll_content(self):
        self.img_lst_v_sizer.Clear(True)

        for lst_1d in self.lst_2d_bmp:
            img_lst_hor_sizer = wx.BoxSizer(wx.HORIZONTAL)

            for i, i_bmp in enumerate(lst_1d):
                local_bitmap = wx.StaticBitmap(self, bitmap=i_bmp)
                if self.parent_panel.local_bbox is None:
                    slice_string = "Slice[" + str(i) + ":" + str(i + 1) + ", :, :]"
                else:
                    bbx = self.parent_panel.local_bbox
                    slice_string = "Image # " + str(bbx[4] + i)

                slice_sub_info_txt = wx.StaticText(self, -1, slice_string)

                sigle_slice_sizer = wx.BoxSizer(wx.VERTICAL)
                sigle_slice_sizer.Clear(True)

                sigle_slice_sizer.Add(
                    local_bitmap, proportion=0, flag=wx.ALIGN_CENTRE | wx.ALL, border=2
                )
                sigle_slice_sizer.Add(
                    slice_sub_info_txt,
                    proportion=0,
                    flag=wx.ALIGN_CENTRE | wx.ALL,
                    border=2,
                )
                img_lst_hor_sizer.Add(
                    sigle_slice_sizer,
                    proportion=0,
                    flag=wx.ALIGN_CENTER | wx.ALL,
                    border=2,
                )

            self.img_lst_v_sizer.Add(img_lst_hor_sizer, 0, wx.CENTER | wx.ALL, 10)

            self.n_img += 1

        self.parent_panel.Pframe.Layout()
        self.SetScrollRate(1, 1)

    def OnMouseMotion(self, event):

        self.Mouse_Pos_x, self.Mouse_Pos_y = event.GetPosition()

    def OnMouseWheel(self, event):
        # Getting amount of scroll steps to do
        sn_mov = math.copysign(1, float(event.GetWheelRotation()))
        self.scroll_rot += sn_mov

        # Getting relative position of scrolling area to keep after scrolling
        v_size_x, v_size_y = self.GetVirtualSize()
        self.Mouse_Pos_x, self.Mouse_Pos_y = event.GetPosition()

        View_start_x, View_start_y = self.GetViewStart()

        self.x_uni = float(View_start_x + self.Mouse_Pos_x) / float(v_size_x)
        self.y_uni = float(View_start_y + self.Mouse_Pos_y) / float(v_size_y)

    def img_refresh(self, i_bmp_new):

        self.lst_2d_bmp = i_bmp_new
        self.set_scroll_content()

    def OnIdle(self, event):
        if self.scroll_rot != 0:

            self.SetScrollRate(1, 1)

            self.parent_panel.to_re_zoom(self.scroll_rot)
            self.scroll_rot = 0
            v_size_x, v_size_y = self.GetVirtualSize()
            new_scroll_pos_x = float(self.x_uni * v_size_x - self.Mouse_Pos_x)
            new_scroll_pos_y = float(self.y_uni * v_size_y - self.Mouse_Pos_y)

            self.Scroll(new_scroll_pos_x, new_scroll_pos_y)

        else:
            if self.old_Mouse_Pos_x != -1 and self.old_Mouse_Pos_y != -1:
                # print "need to move IMG"
                pass

            self.old_Mouse_Pos_x = self.Mouse_Pos_x
            self.old_Mouse_Pos_y = self.Mouse_Pos_y


class buttons_panel(wx.Panel):
    def __init__(self, outer_panel):
        super().__init__(outer_panel)
        self.parent_panel = outer_panel

        Show_Its_CheckBox = wx.CheckBox(self, -1, "Show I nums")
        Show_Its_CheckBox.SetValue(True)
        Show_Its_CheckBox.Bind(wx.EVT_CHECKBOX, self.OnItsCheckbox)

        self.my_sizer = wx.BoxSizer(wx.VERTICAL)
        self.my_sizer.Add(Show_Its_CheckBox, proportion=0, flag=wx.ALIGN_TOP, border=5)

        if self.parent_panel.segn_lst_in is not None:
            Show_Msk_CheckBox = wx.CheckBox(self, -1, "Show Mask")
            Show_Msk_CheckBox.SetValue(True)
            Show_Msk_CheckBox.Bind(wx.EVT_CHECKBOX, self.OnMskCheckbox)

            self.my_sizer.Add(
                Show_Msk_CheckBox, proportion=0, flag=wx.ALIGN_TOP, border=5
            )

            masck_conv_str = "\n Mask Convention:"
            masck_conv_str += "\n [Valid]                      =  \\\\\\\\\\\\  "
            masck_conv_str += "\n [Foreground]           =  //////  "
            masck_conv_str += "\n [Background]          =  ||||||  "
            masck_conv_str += "\n [BackgroundUsed]  =  ------"

            label_mask = wx.StaticText(self, -1, masck_conv_str)
            self.my_sizer.Add(label_mask, proportion=0, flag=wx.ALIGN_TOP, border=5)

        label_palette = wx.StaticText(self, -1, "\nColour Palettes")

        self.RadButtb2w = wx.RadioButton(self, -1, "black2white")
        self.RadButtw2b = wx.RadioButton(self, -1, "white2black")
        self.RadButtha = wx.RadioButton(self, -1, "hot ascend")
        self.RadButthd = wx.RadioButton(self, -1, "hot descend")

        self.RadButtb2w.Bind(wx.EVT_RADIOBUTTON, self.OnButtb2w)
        self.RadButtw2b.Bind(wx.EVT_RADIOBUTTON, self.OnButtw2b)
        self.RadButtha.Bind(wx.EVT_RADIOBUTTON, self.OnButtha)
        self.RadButthd.Bind(wx.EVT_RADIOBUTTON, self.OnButthd)

        self.my_sizer.Add(label_palette, proportion=0, flag=wx.ALIGN_TOP, border=5)

        self.my_sizer.Add(self.RadButtb2w, proportion=0, flag=wx.ALIGN_TOP, border=5)
        self.my_sizer.Add(self.RadButtw2b, proportion=0, flag=wx.ALIGN_TOP, border=5)
        self.my_sizer.Add(self.RadButtha, proportion=0, flag=wx.ALIGN_TOP, border=5)
        self.my_sizer.Add(self.RadButthd, proportion=0, flag=wx.ALIGN_TOP, border=5)

        self.my_sizer.SetMinSize((50, 300))
        self.SetSizer(self.my_sizer)

    def OnButtb2w(self, event):
        self.OnButtUpdate()

    def OnButtw2b(self, event):
        self.OnButtUpdate()

    def OnButtha(self, event):
        self.OnButtUpdate()

    def OnButthd(self, event):
        self.OnButtUpdate()

    def OnButtUpdate(self):
        print("OnButtUpdate(self):")
        if self.RadButtb2w.GetValue():
            self.parent_panel.to_change_palette("black2white")
        elif self.RadButtw2b.GetValue():
            self.parent_panel.to_change_palette("white2black")
        elif self.RadButtha.GetValue():
            self.parent_panel.to_change_palette("hot ascend")
        else:
            self.parent_panel.to_change_palette("hot descend")

    def OnItsCheckbox(self, event):

        if event.IsChecked():
            self.parent_panel.to_show_nums()
        else:
            self.parent_panel.to_hide_nums()

    def OnMskCheckbox(self, event):
        if event.IsChecked():
            self.parent_panel.to_show_mask()
        else:
            self.parent_panel.to_hide_mask()


def bigger_size(str_label, lst_col):
    lng_label_ini = len(str_label)
    lng_lst_col = len(lst_col)
    lng_final = lng_label_ini

    if lng_lst_col < 30:
        pos_lst = list(range(lng_lst_col))

    else:
        pos_lst = list(range(15))
        pos_lst += list(range(lng_lst_col - 15, lng_lst_col))

    lng_cel_zero = 0
    for pos in pos_lst:
        lng_cel_pos = len(str(lst_col[pos]))
        if lng_cel_pos > lng_cel_zero:
            lng_cel_zero = lng_cel_pos

        if lng_cel_zero > lng_label_ini:
            lng_final = lng_cel_zero

    return lng_final



from __future__ import annotations

import copy

import pytest

from dxtbx.serialize import load

from dials.algorithms.profile_model.ellipsoid.algorithm import run_ellipsoid_refinement
from dials.algorithms.profile_model.ellipsoid.indexer import reindex
from dials.array_family import flex


@pytest.mark.xdist_group(name="group1")
def test_run_ellipsoid_refinement(dials_data):
    # Download data set and the internally referenced images
    ssx = dials_data("cunir_serial_processed", pathlib=True)
    dials_data("cunir_serial", pathlib=True)

    refls = flex.reflection_table.from_file(ssx / "indexed.refl")
    expts = load.experiment_list(ssx / "indexed.expt", check_format=False)[0:1]
    refls = refls.select_on_experiment_identifiers([expts[0].identifier])
    refls = reindex(refls, expts[0])

    e_angular2 = {"radial": 0.0179, "angular": 0.0148}
    e_angular4 = {"radial": 0.0179, "angular_0": 0.0207, "angular_1": 0.0029}
    e_simple1 = {"spherical": 0.0159}
    e_simple6 = {"min": 0.0034, "mid": 0.0113, "max": 0.02520}

    initial_crystal = copy.deepcopy(expts[0].crystal)
    for model, expected in zip(
        ["angular2", "angular4", "simple1", "simple6"],
        [e_angular2, e_angular4, e_simple1, e_simple6],
    ):
        elist = copy.deepcopy(expts[0:1])
        out_expt, _, __ = run_ellipsoid_refinement(
            elist, refls, 0.00062, profile_model=model
        )
        for k, v in out_expt.profiles()[0].mosaicity().items():
            assert expected[k] == pytest.approx(v, abs=1e-4)
        assert (
            out_expt[0].crystal.get_unit_cell().parameters()
            != initial_crystal.get_unit_cell().parameters()
        )
        assert list(out_expt[0].crystal.get_A()) != list(initial_crystal.get_A())
        del elist[0].crystal.mosaicity

    # now try fix uc
    elist = copy.deepcopy(expts[0:1])
    out_expt, _, __ = run_ellipsoid_refinement(
        elist,
        refls,
        0.00062,
        profile_model="angular4",
        fix_orientation=True,
        fix_unit_cell=True,
    )
    e_angular4 = {"radial": 0.0182, "angular_0": 0.0207, "angular_1": 0.0052}
    for k, v in out_expt.profiles()[0].mosaicity().items():
        assert e_angular4[k] == pytest.approx(v, abs=1e-4)
    assert out_expt[0].crystal.get_unit_cell().parameters() == pytest.approx(
        initial_crystal.get_unit_cell().parameters(), abs=1e-12
    )
    assert list(out_expt[0].crystal.get_A()) == pytest.approx(
        list(initial_crystal.get_A()), abs=1e-12
    )



from __future__ import annotations

import collections
import math
import os
import pickle

import pytest

from dxtbx.model.experiment_list import ExperimentListFactory

from dials.array_family import flex


@pytest.fixture(scope="module")
def data(dials_regression):  # read experiments and reflections
    directory = os.path.join(dials_regression, "integration_test_data", "shoeboxes")
    experiments_filename = os.path.join(directory, "integrated_experiments.json")
    reflections_filename = os.path.join(directory, "shoeboxes_0_0.pickle")
    reference_filename = os.path.join(directory, "reference_profiles.pickle")

    experiments = ExperimentListFactory.from_json_file(
        experiments_filename, check_format=False
    )
    reflections = flex.reflection_table.from_file(reflections_filename)
    with open(reference_filename, "rb") as fh:
        reference = pickle.load(fh, encoding="bytes")

    Data = collections.namedtuple("Data", ["experiments", "reflections", "reference"])
    return Data(experiments=experiments, reflections=reflections, reference=reference)


def test_gaussianrs_mask_calculator(data):
    from dials.algorithms.integration.parallel_integrator import MaskCalculatorFactory

    algorithm = MaskCalculatorFactory.create(data.experiments)
    reflections = flex.reflection_table_to_list_of_reflections(data.reflections)

    for r in reflections:
        algorithm(r, False)


def test_simple_background_calculator(data):
    from dials.algorithms.background.simple.algorithm import (
        SimpleBackgroundCalculatorFactory,
    )

    algorithm = SimpleBackgroundCalculatorFactory.create(data.experiments)
    reflections = flex.reflection_table_to_list_of_reflections(data.reflections)

    errors = 0
    for r in reflections:
        try:
            algorithm(r)
        except Exception:
            errors += 1
    assert len(reflections) == 15193
    assert errors == 333


def test_glm_background_calculator(data):
    from dials.algorithms.background.glm.algorithm import GLMBackgroundCalculatorFactory

    algorithm = GLMBackgroundCalculatorFactory.create(data.experiments)
    reflections = flex.reflection_table_to_list_of_reflections(data.reflections)

    errors = 0
    for r in reflections:
        try:
            algorithm(r)
        except Exception:
            errors += 1
    assert len(reflections) == 15193
    assert errors == 333


class IntensityCalculatorFactory:
    @staticmethod
    def create(data, detector_space=False, deconvolution=False):
        from dials.algorithms.integration.parallel_integrator import (
            GaussianRSMultiCrystalReferenceProfileData,
            GaussianRSReferenceProfileData,
            ReferenceProfileData,
        )
        from dials.algorithms.profile_model.gaussian_rs.algorithm import (
            GaussianRSIntensityCalculatorFactory,
        )
        from dials.algorithms.profile_model.gaussian_rs.transform import TransformSpec
        from dials.algorithms.profile_model.modeller import CircleSampler

        reference = data.reference[0]
        experiments = data.experiments

        assert len(reference) % 9 == 0
        num_scan_points = len(reference) // 9

        data_spec = GaussianRSMultiCrystalReferenceProfileData()
        for e in experiments:

            sampler = CircleSampler(
                e.detector[0].get_image_size(),
                e.scan.get_array_range(),
                num_scan_points,
            )

            spec = TransformSpec(
                e.beam,
                e.detector,
                e.goniometer,
                e.scan,
                e.profile.sigma_b(deg=False),
                e.profile.sigma_m(deg=False),
                e.profile.n_sigma() * 1.5,
                5,
            )

            temp = reference

            reference = ReferenceProfileData()
            for d, m in temp:
                reference.append(d, m)

            spec = GaussianRSReferenceProfileData(reference, sampler, spec)

            data_spec.append(spec)

        return GaussianRSIntensityCalculatorFactory.create(
            data_spec, detector_space, deconvolution
        )


def test_gaussianrs_reciprocal_space_intensity_calculator(data):
    algorithm = IntensityCalculatorFactory.create(
        data, detector_space=False, deconvolution=False
    )

    reflections = flex.reflection_table_to_list_of_reflections(data.reflections)

    errors = 0
    for r in reflections:
        try:
            algorithm(r, [])
        except Exception:
            errors += 1

    assert len(reflections) == 15193
    assert errors == 5296


def test_gaussianrs_detector_space_intensity_calculator(data):
    algorithm = IntensityCalculatorFactory.create(
        data, detector_space=True, deconvolution=False
    )

    reflections = flex.reflection_table_to_list_of_reflections(data.reflections)

    errors = 0
    for r in reflections:
        try:
            algorithm(r, [])
            partiality_old = r.get("partiality_old")
            partiality_new = r.get("partiality")
        except Exception as e:
            if ("Partiality too small to fit" in str(e)) or (
                "DIALS_ASSERT(fit.niter() < 100) failure." in str(e)
            ):
                errors += 1
                continue
            raise

        assert partiality_old < 1.0 and partiality_old >= 0
        assert partiality_new < 1.0 and partiality_new >= 0

    assert len(reflections) == 15193
    assert errors == 4802


def test_gaussianrs_detector_space_with_deconvolution_intensity_calculator(data):

    algorithm = IntensityCalculatorFactory.create(
        data, detector_space=True, deconvolution=True
    )

    reflections = flex.reflection_table_to_list_of_reflections(data.reflections)

    errors = 0
    for r in reflections:
        try:
            algorithm(r, [])
            partiality_old = r.get("partiality_old")
            partiality_new = r.get("partiality")
        except Exception as e:
            if ("Partiality too small to fit" in str(e)) or (
                "DIALS_ASSERT(fit.niter() < 100) failure." in str(e)
            ):
                errors += 1
                continue
            raise

        assert partiality_old < 1.0 and partiality_old >= 0
        assert partiality_new < 1.0 and partiality_new >= 0

    assert len(reflections) == 15193
    assert errors == 4802


def test_gaussianrs_detector_space_with_deconvolution_intensity_calculator2(data):
    from scitbx import matrix

    reflections = flex.reflection_table_to_list_of_reflections(data.reflections)

    R = None
    for r in reflections:
        if r.get("partiality") > 0.9 and r.get("intensity.sum.value") > 100:
            R = r
            break
    assert R is not None

    px = R.get("xyzcal.px")
    mm = R.get("xyzcal.mm")

    px1 = (px[0] - 3, px[1] - 3, px[2])
    px2 = (px[0] + 3, px[1] + 3, px[2])
    mm1 = (mm[0] - 3 * 0.172, mm[1] - 3 * 0.172, mm[2])
    mm2 = (mm[0] + 3 * 0.172, mm[1] + 3 * 0.172, mm[2])

    s11 = matrix.col(
        data.experiments[0].detector[0].get_pixel_lab_coord(px1[0:2])
    ).normalize()
    s12 = matrix.col(
        data.experiments[0].detector[0].get_pixel_lab_coord(px2[0:2])
    ).normalize()

    R1 = R.copy()
    R2 = R.copy()

    R1.set_vec3_double("xyzcal.px", px1)
    R1.set_vec3_double("xyzcal.mm", mm1)
    R1.set_vec3_double("s1", s11)

    R2.set_vec3_double("xyzcal.px", px2)
    R2.set_vec3_double("xyzcal.mm", mm2)
    R2.set_vec3_double("s1", s12)

    compute_intensity = IntensityCalculatorFactory.create(
        data, detector_space=True, deconvolution=False
    )

    compute_intensity(R, [])
    R.set_double("intensity.prf.value.old", R.get("intensity.prf.value"))
    R.set_double("intensity.prf.variance.old", R.get("intensity.prf.variance"))

    R_no_deconvolution = R.copy()
    # print "Partiality", R.get("partiality")
    # print "Partiality Old", R.get("partiality_old")
    # print "Intensity", R.get("intensity.prf.value")
    # print "Intensity Old", R.get("intensity.prf.value.old")

    from dials.algorithms.integration.parallel_integrator import MaskCalculatorFactory

    mask_calculator = MaskCalculatorFactory.create(data.experiments)

    mask_calculator(R1, True)
    mask_calculator(R2, True)

    # print R.get("shoebox").mask.as_numpy_array()[0,:,:]

    compute_intensity = IntensityCalculatorFactory.create(
        data, detector_space=True, deconvolution=True
    )

    compute_intensity(R, [R1, R2])

    R_deconvolution = R.copy()

    partiality1 = R_no_deconvolution.get("partiality")
    partiality2 = R_deconvolution.get("partiality")
    intensity = R_deconvolution.get("intensity.prf.value")
    variance = R_deconvolution.get("intensity.prf.variance")

    assert partiality1 <= partiality2

    assert abs(intensity - 179.05997142249996) < 1e-7
    assert abs(variance - 203.56992949599677) < 1e-7

    # print "Partiality", R.get("partiality")
    # print "Partiality Old", R.get("partiality_old")
    # print "Intensity", R.get("intensity.prf.value")
    # print "Intensity Old", R.get("intensity.prf.value.old")
    # print "Variance", R.get("intensity.prf.variance")
    # print "Variance Old", R.get("intensity.prf.variance.old")

    # print R.get("shoebox").mask.all()
    # sbox = R.get("shoebox")
    # print sbox.mask.count(5) + sbox.mask.count(37) + sbox.mask.count(51)


def test_gaussianrs_profile_data_pickling(data):
    from dials.algorithms.integration.parallel_integrator import (
        GaussianRSMultiCrystalReferenceProfileData,
        GaussianRSReferenceProfileData,
        ReferenceProfileData,
    )
    from dials.algorithms.profile_model.gaussian_rs.transform import TransformSpec
    from dials.algorithms.profile_model.modeller import CircleSampler

    reference = data.reference[0]
    experiments = data.experiments

    assert len(reference) % 9 == 0
    num_scan_points = len(reference) // 9

    data_spec = GaussianRSMultiCrystalReferenceProfileData()
    for e in experiments:

        sampler = CircleSampler(
            e.detector[0].get_image_size(), e.scan.get_array_range(), num_scan_points
        )

        spec = TransformSpec(
            e.beam,
            e.detector,
            e.goniometer,
            e.scan,
            e.profile.sigma_b(deg=False),
            e.profile.sigma_m(deg=False),
            e.profile.n_sigma() * 1.5,
            5,
        )

        temp = reference

        reference = ReferenceProfileData()
        for d, m in temp:
            reference.append(d, m)

        spec = GaussianRSReferenceProfileData(reference, sampler, spec)

        data_spec.append(spec)

    s = pickle.dumps(data_spec)

    pickle.loads(s)


def test_gaussianrs_reference_profile_calculator(data):
    from dials.algorithms.profile_model.gaussian_rs.algorithm import (
        GaussianRSReferenceCalculatorFactory,
    )

    algorithm = GaussianRSReferenceCalculatorFactory.create(data.experiments)
    reflections = flex.reflection_table_to_list_of_reflections(data.reflections)

    errors = 0
    for r in reflections:
        try:
            algorithm(r)
        except Exception:
            errors += 1
    assert len(reflections) == 15193
    assert errors == 0

    profiles = algorithm.reference_profiles()

    count = 0
    for i in range(len(profiles)):
        p = profiles[i].reference()
        for j in range(len(p)):
            d = p.data(j)
            p.mask(j)
            if len(d) != 0:
                count += 1

    assert count == 9


def test_job_list():
    from dials.algorithms.integration.parallel_integrator import SimpleBlockList

    jobs = SimpleBlockList((0, 60), 20)

    assert jobs[0] == (0, 20)
    assert jobs[1] == (10, 30)
    assert jobs[2] == (20, 40)
    assert jobs[3] == (30, 50)
    assert jobs[4] == (40, 60)

    for frame in range(0, 15):
        assert jobs.block_index(frame) == 0
    for frame in range(16, 25):
        assert jobs.block_index(frame) == 1
    for frame in range(26, 35):
        assert jobs.block_index(frame) == 2
    for frame in range(36, 45):
        assert jobs.block_index(frame) == 3
    for frame in range(46, 60):
        assert jobs.block_index(frame) == 4


def test_reflection_manager(data):
    from dials.algorithms.integration.parallel_integrator import (
        SimpleBlockList,
        SimpleReflectionManager,
    )

    jobs = SimpleBlockList((0, 60), 20)
    reflections = data.reflections

    manager = SimpleReflectionManager(jobs, reflections, 5)

    def check_job(index):
        r = manager.split(index)
        selection = r.get_flags(r.flags.dont_integrate)
        r1 = r.select(~selection)

        j0, j1 = jobs[index]

        for rr in r1.rows():
            z0, z1 = rr["bbox"][4:6]
            zc = int(math.floor((z0 + z1) / 2.0))
            j = jobs.block_index(zc)
            assert j == index
            assert z0 >= j0 and z1 <= j1

    check_job(0)
    check_job(1)
    check_job(2)
    check_job(3)
    check_job(4)



from __future__ import annotations


class NullBackgroundExt:
    """An extension class implementing Null background subtraction."""

    name = "null"

    def __init__(self, params, experiments):
        """Initialise the algorithm.

        :param params: The input phil parameters
        :param experiments: The experiment list
        """
        pass

    def compute_background(self, reflections, image_volume=None):
        """
        Compute the background.

        :param reflections: The list of reflections
        """
        from dials.algorithms.background import set_shoebox_background_value
        from dials.array_family import flex

        set_shoebox_background_value(reflections["shoebox"], 0)
        return flex.bool(len(reflections), True)



from __future__ import annotations

import logging
import math

import iotbx.phil
from cctbx import sgtbx
from cctbx.crystal import symmetry as crystal_symmetry
from cctbx.miller import set as miller_set
from cctbx.sgtbx import space_group as sgtbx_space_group
from libtbx.utils import format_float_with_standard_uncertainty

from dials.algorithms.symmetry import origin
from dials.array_family import flex
from dials.util import log, show_mail_handle_errors
from dials.util.options import ArgumentParser, reflections_and_experiments_from_files
from dials.util.version import dials_version

logger = logging.getLogger("dials.command_line.check_indexing_symmetry")

help_message = """

This program can be used to analyse the correlation coefficients between
reflections related by the symmetry operators belonging to the space group of
the input experiment.expt file. It can also check for misindexing of
the diffraction pattern, possibly as a result of an incorrect beam centre.

Examples::

  dials.check_indexing_symmetry indexed.expt indexed.refl \\
    grid=1 symop_threshold=0.7


  dials.check_indexing_symmetry indexed.expt indexed.refl \\
    grid_l=3 symop_threshold=0.7
"""

phil_scope = iotbx.phil.parse(
    """
d_min = 0
  .type = float
  .help = "High resolution limit to use for analysis"
d_max = 0
  .type = float
  .help = "Low resolution limit to use for analysis"
symop_threshold = 0
  .type = float
  .help = "Threshold above which we consider a symmetry operator true"
          "at approximately 95% confidence."
grid = 0
  .type = int
  .help = "Search scope for testing misindexing on h, k, l."
grid_h = 0
  .type = int
  .help = "Search scope for testing misindexing on h."
grid_k = 0
  .type = int
  .help = "Search scope for testing misindexing on k."
grid_l = 0
  .type = int
  .help = "Search scope for testing misindexing on l."
asu = False
  .type = bool
  .help = "Perform search comparing within ASU (assumes input symm)"
normalise = False
  .type = bool
  .help = "Normalise intensities before calculating correlation coefficients."
normalise_bins = 0
  .type = int
  .help = "Number of resolution bins for normalisation"
reference = None
  .type = path
  .help = "Correctly indexed reference set for comparison"
output {
  log = dials.check_indexing_symmetry.log
    .type = str
}
""",
    process_includes=True,
)


def standard_error_of_pearson_cc(corr_coeffs, n):
    """Calculate the standard error of Pearson's correlation coefficient.

    This uses the formula SE = sqrt((1 - r^2) / (n - 2)), which is correct
    only if the data from which the CCs were calculated are normally
    distributed.

    Args:
        corr_coeffs: a flex.double array of correlation coefficient values
        n: a flex.int array of sample sizes used to calculate corr_coeffs

    Returns:
        flex.double: the standard error for values in corr_coeffs
    """

    numerator = 1 - flex.pow2(corr_coeffs)
    denominator = n - 2
    sel = denominator < 1
    denominator.set_selected(sel, 1)
    vals = numerator / denominator.as_double()
    vals.set_selected(sel, float("NaN"))
    return vals


def format_cc_with_standard_error(cc, se, decimal_places=3):
    if math.isnan(se):
        return "?"
    minimum = pow(0.1, decimal_places)
    if se >= minimum:
        cc_str = format_float_with_standard_uncertainty(cc, se, minimum=1.0e-3)
    else:
        cc_str = "{val:0.{dp}f}".format(val=cc, dp=decimal_places)
    return cc_str


def get_symop_correlation_coefficients(miller_array, use_binning=False):
    corr_coeffs = flex.double()
    n_refs = flex.int()
    space_group = miller_array.space_group()
    for smx in space_group.smx():
        reindexed_array = miller_array.change_basis(sgtbx.change_of_basis_op(smx))
        intensity, intensity_rdx = reindexed_array.common_sets(miller_array)
        if use_binning:
            intensity.use_binning_of(miller_array)
            intensity_rdx.use_binning_of(miller_array)
            cc = intensity.correlation(intensity_rdx, use_binning=use_binning)
            corr_coeffs.append(
                flex.mean_weighted(
                    flex.double(i for i in cc.data if i is not None),
                    flex.double(
                        j for i, j in zip(cc.data, cc.binner.counts()) if i is not None
                    ),
                )
            )
        else:
            corr_coeffs.append(
                intensity.correlation(
                    intensity_rdx, use_binning=use_binning
                ).coefficient()
            )
        n_refs.append(intensity.size())
    return corr_coeffs, n_refs


def normalise_intensities(miller_array, n_bins=10):
    miller_array.setup_binner(n_bins=n_bins)
    nomalisations = miller_array.amplitude_quasi_normalisations()
    miller_array = miller_array.customized_copy(
        data=miller_array.data() / nomalisations.data()
    )
    return miller_array


def test_crystal_pointgroup_symmetry(reflections, experiment, params):
    crystal = experiment.crystal

    # in case we pass in reflections from integration
    reflections = reflections.select(reflections["intensity.sum.variance"] > 0)
    reflections = reflections.select(reflections["intensity.sum.value"] > 0)
    original_miller_indices = reflections["miller_index"]

    space_group = crystal.get_space_group()
    unit_cell = crystal.get_unit_cell()

    cs = crystal_symmetry(unit_cell, space_group.type().lookup_symbol())

    ms = miller_set(cs, original_miller_indices)
    ms = ms.array(
        reflections["intensity.sum.value"]
        / flex.sqrt(reflections["intensity.sum.variance"])
    )

    if params.d_min or params.d_max:
        d_spacings = ms.d_spacings().data()
        sel = d_spacings >= params.d_min
        if params.d_max > 0:
            sel = sel & (d_spacings <= params.d_max)
        ms = ms.select(sel)

    if params.normalise:
        if params.normalise_bins:
            ms = normalise_intensities(ms, n_bins=params.normalise_bins)
        else:
            ms = normalise_intensities(ms)

    logger.info("Check symmetry operations on %d reflections:", ms.size())
    logger.info("")
    logger.info("%20s %6s %5s", "Symop", "Nref", "CC")

    true_symops = []

    ccs, n_refs = get_symop_correlation_coefficients(ms)
    ses = standard_error_of_pearson_cc(ccs, n_refs)

    for smx, cc, n_ref, se in zip(space_group.smx(), ccs, n_refs, ses):
        accept = ""
        if params.symop_threshold:
            if (cc - 2.0 * se) > params.symop_threshold:
                true_symops.append(smx)
                accept = "***"
        cc_str = format_cc_with_standard_error(cc, se)
        logger.info("%20s %6d %s %s", smx, n_ref, cc_str, accept)

    if params.symop_threshold:
        sg = sgtbx_space_group()
        for symop in true_symops:
            sg = sg.expand_smx(symop)
        for ltr in space_group.ltr():
            sg = sg.expand_ltr(ltr)
        sg_symbols = sg.match_tabulated_settings()
        logger.info("")
        logger.info(
            "Derived point group from symmetry operations: %s",
            sg_symbols.hermann_mauguin(),
        )
        logger.info("")

    return


def offset_miller_indices(miller_indices, offset):
    return flex.miller_index(
        *[mi.iround() for mi in (miller_indices.as_vec3_double() + offset).parts()]
    )


def get_indexing_offset_correlation_coefficients(
    reflections,
    crystal,
    grid,
    d_min=None,
    d_max=None,
    map_to_asu=False,
    grid_h=0,
    grid_k=0,
    grid_l=0,
    reference=None,
):
    if grid:
        if grid_h == 0:
            grid_h = grid
        if grid_k == 0:
            grid_k = grid
        if grid_l == 0:
            grid_l = grid

    return origin.get_hkl_offset_correlation_coefficients(
        reflections,
        crystal,
        map_to_asu=map_to_asu,
        grid_h=grid_h,
        grid_k=grid_k,
        grid_l=grid_l,
        reference=reference,
    )


def test_P1_crystal_indexing(reflections, experiment, params):
    if not (params.grid or params.grid_h or params.grid_k or params.grid_l):
        return

    logger.info("Checking HKL origin:")
    logger.info("")
    logger.info("dH dK dL %6s %5s", "Nref", "CC")

    if params.reference:
        reference = flex.reflection_table.from_file(params.reference)
    else:
        reference = None

    offsets, ccs, n_refs = get_indexing_offset_correlation_coefficients(
        reflections,
        experiment.crystal,
        grid=params.grid,
        d_min=params.d_min,
        d_max=params.d_max,
        map_to_asu=params.asu,
        grid_h=params.grid_h,
        grid_k=params.grid_k,
        grid_l=params.grid_l,
        reference=reference,
    )

    ses = standard_error_of_pearson_cc(ccs, n_refs)

    for (h, k, l), cc, n, se in zip(offsets, ccs, n_refs, ses):
        if cc > params.symop_threshold or (h == k == l == 0):
            cc_str = format_cc_with_standard_error(cc, se)
            logger.info("%2d %2d %2d %6d %s", h, k, l, n, cc_str)

    logger.info("")

    return


@show_mail_handle_errors()
def run(args=None):
    usage = "dials.check_indexing_symmetry [options] indexed.expt indexed.refl"

    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_reflections=True,
        read_experiments=True,
        check_format=False,
        epilog=help_message,
    )

    params, options = parser.parse_args(args, show_diff_phil=True)

    # Configure the logging
    log.config(logfile=params.output.log)
    logger.info(dials_version())

    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )
    if len(reflections) == 0 or len(experiments) == 0:
        parser.print_help()
        return
    assert len(reflections) == 1
    assert len(experiments) == 1
    experiment = experiments[0]
    reflections = reflections[0]

    # remove reflections with 0, 0, 0 index
    zero = reflections["miller_index"] == (0, 0, 0)
    logger.info("Removing %d unindexed reflections", zero.count(True))
    reflections = reflections.select(~zero)

    h, k, l = reflections["miller_index"].as_vec3_double().parts()

    h = h.iround()
    k = k.iround()
    l = l.iround()

    logger.info("Range on h: %d to %d", flex.min(h), flex.max(h))
    logger.info("Range on k: %d to %d", flex.min(k), flex.max(k))
    logger.info("Range on l: %d to %d", flex.min(l), flex.max(l))

    test_P1_crystal_indexing(reflections, experiment, params)
    test_crystal_pointgroup_symmetry(reflections, experiment, params)


if __name__ == "__main__":
    run()



# DIALS_ENABLE_COMMAND_LINE_COMPLETION
"""
Refine the diffraction geometry of input experiments against the input indexed
reflections. For rotation scans, the model may be either static (the same for
all reflections) or scan-varying (dependent on image number in the scan).
Other basic parameters include control over output filenames, fixing of
certain parameters of each model and options that control the number of
reflections used in refinement.

Examples::

  dials.refine indexed.expt indexed.refl

  dials.refine indexed.expt indexed.refl scan_varying=(False/True/Auto)
"""


from __future__ import annotations

import copy
import logging
import sys

import libtbx.phil
from libtbx import Auto

import dials.util
import dials.util.log
from dials.algorithms.refinement import (
    DialsRefineConfigError,
    DialsRefineRuntimeError,
    RefinerFactory,
)
from dials.algorithms.refinement.corrgram import create_correlation_plots
from dials.array_family import flex
from dials.util.options import ArgumentParser, reflections_and_experiments_from_files
from dials.util.version import dials_version

logger = logging.getLogger("dials.command_line.refine")

# The phil scope
phil_scope = libtbx.phil.parse(
    """

  output {
    experiments = refined.expt
      .type = str
      .help = "The filename for refined experimental models"

    reflections = refined.refl
      .type = str
      .help = "The filename for reflections with updated predictions"

    include_unused_reflections = True
      .type = bool
      .help = "If True, keep reflections unused in refinement in updated"
              "reflections file. Otherwise, remove them"
      .expert_level = 1

    matches = None
      .type = str
      .help = "The filename for output of the reflection table for reflections"
              "used in refinement, containing extra columns used internally."
              "Intended for debugging purposes only"
      .expert_level = 2

    centroids = None
      .type = str
      .help = "The filename for the table of centroids at the end of"
              "refinement"
      .expert_level = 1

    parameter_table = None
      .type = str
      .help = "The filename for the table of scan varying parameter values"
      .expert_level = 1

    log = dials.refine.log
      .type = str

    include scope dials.algorithms.refinement.corrgram.phil_scope

    history = None
      .type = str
      .help = "The filename for output of the refinement history json"
      .expert_level = 1
  }

  n_static_macrocycles = 1
    .type = int(value_min=1)
    .help = "Number of macro-cycles of static refinement to perform"

  include scope dials.algorithms.refinement.refiner.phil_scope
""",
    process_includes=True,
)

# local overrides for refiner.phil_scope
phil_overrides = libtbx.phil.parse(
    """
  refinement
  {
    parameterisation.scan_varying = Auto

    reflections.outlier.nproc = Auto
  }
"""
)

working_phil = phil_scope.fetch(sources=[phil_overrides])


def write_centroids_table(refiner, filename):
    """Write a table of centroids from refinement for debugging.

    Args:
        refiner: A dials.algorithms.refinement.refiner.Refiner.
        filename (str): The name of the file to which to write.

    Returns:
        None
    """

    matches = refiner.get_matches()

    header = "H\tK\tL\tFrame_obs\tX_obs\tY_obs\tPhi_obs\tX_calc\t" "Y_calc\tPhi_calc"
    msg_temp = "%d\t%d\t%d\t%d\t%5.3f\t%5.3f\t%9.6f\t%5.3f\t%5.3f\t%9.6f"
    has_del_psi = "delpsical.rad" in matches
    if has_del_psi:
        header += "\tDelta_Psi"
        msg_temp += "\t%9.6f"
    header += "\n"
    msg_temp += "\n"

    with open(filename, "w") as f:
        f.write(header)

        for m in matches:
            (h, k, l) = m["miller_index"]
            frame = m["xyzobs.px.value"][2]
            x_obs, y_obs, phi_obs = m["xyzobs.mm.value"]
            x_calc, y_calc, phi_calc = m["xyzcal.mm"]
            if has_del_psi:
                del_psi = m["delpsical.rad"]
                msg = msg_temp % (
                    h,
                    k,
                    l,
                    frame,
                    x_obs,
                    y_obs,
                    phi_obs,
                    x_calc,
                    y_calc,
                    phi_calc,
                    del_psi,
                )
            else:
                msg = msg_temp % (
                    h,
                    k,
                    l,
                    frame,
                    x_obs,
                    y_obs,
                    phi_obs,
                    x_calc,
                    y_calc,
                    phi_calc,
                )
            f.write(msg)


def run_macrocycle(params, reflections, experiments):
    """Run one macrocycle of refinement.

    One macrocycle of refinement is run, as specified by the PHIL
    parameters, using the centroids from the supplied reflections
    and the initial experimental geometry taken from experiments.


    Args:
        params: The working PHIL parameters.
        reflections: A reflection table containing observed centroids
        experiments: The initial dxtbx experimental geometry models

    Returns:
        tuple: The Refiner, the reflection table with updated predictions
            and flags, and the refinement history object.
    """
    # Get the refiner
    logger.info("Configuring refiner")
    try:
        refiner = RefinerFactory.from_parameters_data_experiments(
            params, reflections, experiments
        )
    except DialsRefineConfigError as e:
        sys.exit(str(e))

    # Refine the geometry
    nexp = len(experiments)
    if nexp == 1:
        logger.info("Performing refinement of a single Experiment...")
    else:
        logger.info(f"Performing refinement of {nexp} Experiments...")

    # Refine and get the refinement history
    try:
        history = refiner.run()
    except DialsRefineRuntimeError as e:
        sys.exit(str(e))

    # Update predictions for all indexed reflections
    logger.info("Updating predictions for indexed reflections")
    preds = refiner.predict_for_indexed()

    # just copy over the columns of interest or columns that may have been
    # updated, leaving behind things added by e.g. scan-varying refinement
    # such as 'block', the U, B and UB matrices and gradients.
    for key in preds:
        if key in reflections.keys() or key in [
            "s1",
            "xyzcal.mm",
            "xyzcal.px",
            "entering",
            "delpsical.rad",
        ]:
            reflections[key] = preds[key]

    # set refinement flags
    assert len(preds) == len(reflections)
    reflections.unset_flags(
        flex.size_t_range(len(reflections)),
        reflections.flags.excluded_for_refinement
        | reflections.flags.used_in_refinement
        | reflections.flags.centroid_outlier
        | reflections.flags.predicted,
    )
    reflections.set_flags(
        preds.get_flags(preds.flags.excluded_for_refinement),
        reflections.flags.excluded_for_refinement,
    )
    reflections.set_flags(
        preds.get_flags(preds.flags.centroid_outlier),
        reflections.flags.centroid_outlier,
    )
    reflections.set_flags(
        preds.get_flags(preds.flags.used_in_refinement),
        reflections.flags.used_in_refinement,
    )
    reflections.set_flags(
        preds.get_flags(preds.flags.predicted), reflections.flags.predicted
    )

    return refiner, reflections, history


def run_dials_refine(experiments, reflections, params):
    """Functional interface to tasks performed by the program dials.refine.

    This runs refinement according to the PHIL parameters in params, using
    taking the initial experimental geometry from experiments, and the
    observed reflection centroids from reflections.

    Static refinement can be done in a number of macrocycles. By default,
    for scans, this will be followed by a macrocycle of scan-varying
    refinement.


    Args:
        experiments: The initial dxtbx experimental geometry models
        reflections: A reflection table containing observed centroids
        params: The working PHIL parameters

    Returns:
        tuple: The refined experiments, the updated reflection table, the
            Refiner object and the refinement history object.

    """

    # Modify options if necessary
    if params.output.correlation_plot.filename is not None:
        params.refinement.refinery.journal.track_parameter_correlation = True

    # scan_varying=Auto is special to dials.refine. Refiner expects it to be
    # True or False, so catch and set here
    scan_varying = params.refinement.parameterisation.scan_varying
    if scan_varying is Auto:
        params.refinement.parameterisation.scan_varying = False

    # Similarly, keep track of sparse to reset that for scan-varying macrocycle
    sparse = params.refinement.parameterisation.sparse

    if params.n_static_macrocycles == 1:
        refiner, reflections, history = run_macrocycle(params, reflections, experiments)
        experiments = refiner.get_experiments()
    else:
        for i in range(params.n_static_macrocycles):
            logger.info("\nStatic refinement macrocycle %s", i + 1)
            refiner, reflections, history = run_macrocycle(
                params, reflections, experiments
            )
            experiments = refiner.get_experiments()

    # Scan-varying macrocycle, if appropriate
    if scan_varying is Auto and refiner.experiment_type == "scans":
        logger.info("\nScan-varying refinement")
        params.refinement.parameterisation.scan_varying = True
        params.refinement.parameterisation.sparse = sparse
        refiner, reflections, history = run_macrocycle(params, reflections, experiments)
        experiments = refiner.get_experiments()

    return experiments, reflections, refiner, history


@dials.util.show_mail_handle_errors()
def run(args=None, phil=working_phil):
    """
    Set up refinement from command line options, files and PHIL parameters.
    Run refinement and save output files as specified.

    Called when running dials.refine as a command-line program

    Args:
        args (list): Additional command-line arguments
        phil: The working PHIL parameters

    Returns:
        None
    """

    # The script usage
    usage = (
        "usage: dials.refine [options] [param.phil] " "models.expt observations.refl"
    )

    # Create the parser
    parser = ArgumentParser(
        usage=usage,
        phil=phil,
        read_reflections=True,
        read_experiments=True,
        check_format=False,
        epilog=__doc__,
    )

    # Parse the command line
    params, options = parser.parse_args(args=args, show_diff_phil=False)

    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )

    # Configure the logging
    dials.util.log.config(verbosity=options.verbose, logfile=params.output.log)

    # Try to load the models and data
    nexp = len(experiments)
    if nexp == 0 or len(reflections) == 0:
        parser.print_help()
        return
    if len(reflections) > 1:
        sys.exit("Only one reflections list can be imported at present")
    reflections = reflections[0]

    # check input is suitable
    msg = (
        "The supplied reflection table does not have the required data " + "column: {0}"
    )
    for key in ["xyzobs.mm.value", "xyzobs.mm.variance"]:
        if key not in reflections:
            sys.exit(msg.format(key))

    logger.info(dials_version())

    # Log the diff phil
    diff_phil = parser.diff_phil.as_str()
    if diff_phil:
        logger.info("The following parameters have been modified:\n")
        logger.info(diff_phil)

    # Warn about potentially unhelpful options
    if params.refinement.mp.nproc > 1:
        logger.warning(
            "Setting nproc > 1 is only helpful in rare "
            "circumstances. It is not recommended for typical data processing "
            "tasks."
        )

    if params.refinement.parameterisation.scan_varying is not False:
        # duplicate crystal if necessary for scan varying - will need
        # to compare the scans with crystals - if not 1:1 will need to
        # split the crystals

        crystal_has_scan = {}
        for j, e in enumerate(experiments):
            if e.crystal in crystal_has_scan:
                if e.scan is not crystal_has_scan[e.crystal]:
                    logger.info(
                        "Duplicating crystal model for scan-varying refinement of experiment %d",
                        j,
                    )
                    e.crystal = copy.deepcopy(e.crystal)
            else:
                crystal_has_scan[e.crystal] = e.scan

    # Run refinement
    experiments, reflections, refiner, history = run_dials_refine(
        experiments, reflections, params
    )

    # For the usual case of refinement of one crystal, print that model for information
    crystals = experiments.crystals()
    if len(crystals) == 1:
        logger.info("")
        logger.info("Final refined crystal model:")
        logger.info(crystals[0])

    # Write table of centroids to file, if requested
    if params.output.centroids:
        logger.info(f"Writing table of centroids to '{params.output.centroids}'")
        write_centroids_table(refiner, params.output.centroids)

    # Write scan-varying parameters to file, if there were any
    if params.output.parameter_table:
        scans = experiments.scans()
        if len(scans) > 1:
            logger.info(
                "Writing a scan-varying parameter table is only supported "
                "for refinement of a single scan"
            )
        else:
            scan = scans[0]
            text = refiner.get_param_reporter().varying_params_vs_image_number(
                scan.get_array_range()
            )
            if text:
                logger.info(
                    "Writing scan-varying parameter table to %s",
                    params.output.parameter_table,
                )
                f = open(params.output.parameter_table, "w")
                f.write(text)
                f.close()
            else:
                logger.info("No scan-varying parameter table to write")

    # Save matches to file for debugging
    if params.output.matches:
        matches = refiner.get_matches()
        logger.info(
            "Saving matches (use for debugging purposes) to %s", params.output.matches
        )
        matches.as_file(params.output.matches)

    # Create correlation plots
    if params.output.correlation_plot.filename is not None:
        create_correlation_plots(refiner, params.output)

    # Save refinement history
    if params.output.history:
        logger.info(f"Saving refinement step history to {params.output.history}")
        history.to_json_file(params.output.history)

    # Save the refined experiments to file
    output_experiments_filename = params.output.experiments
    logger.info(f"Saving refined experiments to {output_experiments_filename}")
    experiments.as_file(output_experiments_filename)

    # Save reflections with updated predictions. This causes a spike in memory
    # usage (https://github.com/dials/dials/issues/2024), so delete big objects
    # we no longer need first
    if params.output.reflections:
        del experiments
        del refiner
        del history
        logger.info(
            "Saving reflections with updated predictions to %s",
            params.output.reflections,
        )
        if params.output.include_unused_reflections:
            reflections.as_file(params.output.reflections)
        else:
            sel = reflections.get_flags(reflections.flags.used_in_refinement)
            reflections.select(sel).as_file(params.output.reflections)


if __name__ == "__main__":
    run()



"""Versions of refinement classes for two theta refinement of the unit cell"""


from __future__ import annotations

import logging
from math import pi, sqrt

from scitbx import matrix
from scitbx.math import five_number_summary

from dials.algorithms.refinement.parameterisation.prediction_parameters import (
    PredictionParameterisation,
)
from dials.algorithms.refinement.prediction.managed_predictors import (
    ExperimentsPredictor,
)
from dials.algorithms.refinement.reflection_manager import ReflectionManager
from dials.algorithms.refinement.target import Target
from dials.array_family import flex
from dials.util import tabulate

logger = logging.getLogger(__name__)

# constants
RAD2DEG = 180.0 / pi
DEG2RAD = pi / 180.0


class ConstantTwoThetaWeightingStrategy:
    def calculate_weights(self, reflections):

        reflections["2theta.weights"] = flex.double(len(reflections), 1)
        return reflections


def calc_2theta(reflections, experiments):
    """Calculate and return 2theta angles in radians"""

    twotheta = flex.double(len(reflections), 0.0)
    for iexp, exp in enumerate(experiments):
        isel = (reflections["id"] == iexp).iselection()
        sub_ref = reflections.select(isel)
        s0 = matrix.col(exp.beam.get_s0())
        for ipanel in range(len(exp.detector)):
            sel = sub_ref["panel"] == ipanel
            panel_ref = sub_ref.select(sel)
            x, y, phi = panel_ref["xyzobs.mm.value"].parts()
            s1 = exp.detector[ipanel].get_lab_coord(flex.vec2_double(x, y))
            s1 = s1 / s1.norms() * s0.length()

            sub_isel = isel.select(sel)
            twotheta.set_selected(sub_isel, s1.angle(s0))

    return twotheta


class TwoThetaReflectionManager(ReflectionManager):
    _weighting_strategy = ConstantTwoThetaWeightingStrategy()

    def __init__(self, *args, **kwargs):

        # call base __init__
        super().__init__(*args, **kwargs)

        # set observed 2theta angles
        self._reflections["2theta_obs.rad"] = calc_2theta(
            self._reflections, self._experiments
        )

        # placeholder for calculated 2theta angles
        self._reflections["2theta_cal.rad"] = flex.double(len(self._reflections), 0.0)

        return

    def print_stats_on_matches(self):

        l = self.get_matches()
        nref = len(l)
        if nref == 0:
            logger.warning(
                "Unable to calculate summary statistics for zero observations"
            )
            return

        twotheta_resid = l["2theta_resid"]
        w_2theta = l["2theta.weights"]

        msg = (
            f"\nSummary statistics for {nref} observations" + " matched to predictions:"
        )
        header = ["", "Min", "Q1", "Med", "Q3", "Max"]
        rows = []
        row_data = five_number_summary(twotheta_resid)
        rows.append(
            ["2theta_c - 2theta_o (deg)"] + [f"{e * RAD2DEG:.4g}" for e in row_data]
        )
        row_data = five_number_summary(w_2theta)
        rows.append(["2theta weights"] + [f"{e * DEG2RAD ** 2:.4g}" for e in row_data])
        logger.info(msg)
        logger.info(tabulate(rows, header) + "\n")


class TwoThetaExperimentsPredictor(ExperimentsPredictor):
    def _predict_one_experiment(self, experiment, reflections):

        B = flex.mat3_double(len(reflections), experiment.crystal.get_B())
        r0 = B * reflections["miller_index"].as_vec3_double()
        r0len = r0.norms()
        wl = experiment.beam.get_wavelength()

        # 2theta = 2 * arcsin( |r0| / (2 * |s0| ) )
        reflections["2theta_cal.rad"] = 2.0 * flex.asin(0.5 * r0len * wl)

        reflections.set_flags(
            flex.size_t(len(reflections)), reflections.flags.predicted
        )


class TwoThetaTarget(Target):
    _grad_names = ["d2theta_dp"]
    rmsd_names = ["RMSD_2theta"]
    rmsd_units = ["rad"]

    def __init__(
        self, experiments, predictor, reflection_manager, prediction_parameterisation
    ):
        Target.__init__(
            self,
            experiments,
            predictor,
            reflection_manager,
            prediction_parameterisation,
        )

        # set the single cutoff for 2theta residual to essentially zero
        self._binsize_cutoffs = [1.0e-6]

        # predict reflections and finalise reflection manager
        self.predict()
        self._reflection_manager.finalise()

        return

    def predict(self):
        """perform reflection prediction for the working reflections and update the
        reflection manager"""

        # get the matches
        reflections = self._reflection_manager.get_obs()

        # reset the 'use' flag for all observations
        self._reflection_manager.reset_accepted_reflections()

        # set twotheta in place
        self._reflection_predictor(reflections)

        # calculate  residuals
        reflections["2theta_resid"] = (
            reflections["2theta_cal.rad"] - reflections["2theta_obs.rad"]
        )
        reflections["2theta_resid2"] = flex.pow2(reflections["2theta_resid"])

        # set used_in_refinement flag to all those that had predictions
        mask = reflections.get_flags(reflections.flags.predicted)
        reflections.set_flags(mask, reflections.flags.used_in_refinement)

        # collect the matches
        self.update_matches(force=True)

        return

    @staticmethod
    def _extract_residuals_and_weights(matches):

        # return residuals and weights as 1d flex.double vectors
        residuals = matches["2theta_resid"]

        weights = matches["2theta.weights"]

        return residuals, weights

    @staticmethod
    def _extract_squared_residuals(matches):

        residuals2 = matches["2theta_resid2"]

        return residuals2

    def _rmsds_core(self, reflections):
        """calculate unweighted RMSDs for the specified reflections"""

        resid_2theta = flex.sum(reflections["2theta_resid2"])
        n = len(reflections)

        rmsds = (sqrt(resid_2theta / n),)
        return rmsds

    def achieved(self):
        """RMSD criterion for target achieved"""
        r = self._rmsds if self._rmsds else self.rmsds()

        # reset cached rmsds to avoid getting out of step
        self._rmsds = None

        if r[0] < self._binsize_cutoffs[0]:
            return True
        return False


class TwoThetaPredictionParameterisation(PredictionParameterisation):
    _grad_names = ("d2theta_dp",)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # check that only the unit cell is parameterised
        assert not self._detector_parameterisations
        assert not self._beam_parameterisations
        assert not self._xl_orientation_parameterisations
        assert not self._goniometer_parameterisations
        return

    def _local_setup(self, reflections):

        # we want the wavelength
        self._wavelength = 1.0 / self._s0.norms()

        return

    def _xl_unit_cell_derivatives(self, isel, parameterisation=None, reflections=None):

        # Get required data
        h = self._h.select(isel)
        B = self._B.select(isel)
        wl = self._wavelength.select(isel)

        # get derivatives of the B matrix wrt the parameters
        dB_dxluc_p = [
            None if der is None else flex.mat3_double(len(isel), der.elems)
            for der in parameterisation.get_ds_dp(use_none_as_null=True)
        ]

        d2theta_dp = []

        # loop through the parameters
        for der in dB_dxluc_p:

            if der is None:
                d2theta_dp.append(None)
                continue

            r0 = B * h
            dr0 = der * h
            r0len = r0.norms()
            dr0len = dr0.dot(r0) / r0len

            # 2theta = 2 * arcsin( |r0| / (2 * |s0| ) )
            sintheta = 0.5 * r0len * wl
            fac = 1.0 / flex.sqrt(flex.double(len(wl), 1.0) - flex.pow2(sintheta))
            val = fac * wl * dr0len

            d2theta_dp.append(val)

        return d2theta_dp

    def _grads_xl_unit_cell_loop(self, reflections, results, callback=None):
        """Loop over all crystal unit cell parameterisations, calculate gradients
        and extend the results"""

        # loop over the crystal unit cell parameterisations
        for xlucp in self._xl_unit_cell_parameterisations:

            # Determine (sub)set of reflections affected by this parameterisation
            isel = flex.size_t()
            for exp_id in xlucp.get_experiment_ids():
                isel.extend(self._experiment_to_idx[exp_id])

            # Extend derivative vectors for this crystal unit cell parameterisation
            results = self._extend_gradient_vectors(
                results, self._nref, xlucp.num_free(), keys=self._grad_names
            )

            if len(isel) == 0:
                # if no reflections are in this experiment, skip calculation of
                # gradients, but must still process null gradients by a callback
                if callback is not None:
                    for iparam in range(xlucp.num_free()):
                        results[self._iparam] = callback(results[self._iparam])
                        self._iparam += 1
                else:
                    self._iparam += xlucp.num_free()
                continue

            d2theta_dp = self._xl_unit_cell_derivatives(
                isel, parameterisation=xlucp, reflections=reflections
            )

            for d2theta in d2theta_dp:
                if d2theta is not None:
                    results[self._iparam][self._grad_names[0]].set_selected(
                        isel, d2theta
                    )

                # increment the parameter index pointer
                self._iparam += 1

        return results



from __future__ import annotations

import logging
import math

logger = logging.getLogger(__name__)


def write_background_file(file_name, imageset, n_bins):
    from dials.command_line.background import background

    d, I, sig = background(imageset, imageset.indices()[0], n_bins=n_bins)

    logger.info("Saving background file to %s", file_name)
    with open(file_name, "w") as f:
        for d_, I_, sig_ in zip(d, I, sig):
            f.write(f"{d_:10.4f} {I_:10.2f} {sig_:10.2f}\n")


def write_integrated_hkl(prefix, reflections):
    from scitbx.array_family import flex

    expt_ids = reflections["id"]
    integrated_sel = reflections.get_flags(reflections.flags.integrated_sum)
    for i_expt in range(flex.max(expt_ids) + 1):
        integrated = reflections.select((reflections["id"] == i_expt) & integrated_sel)
        integrated.sort("miller_index")
        h, k, l = integrated["miller_index"].as_vec3_double().parts()
        I = integrated["intensity.sum.value"]
        sigI = flex.sqrt(integrated["intensity.sum.variance"])

        suffix = ""
        if flex.max(expt_ids) > 0:
            suffix = "%i" % (i_expt + 1)
        file_name = f"{prefix}{suffix}.hkl"
        logger.info("Saving reflections to %s", file_name)
        with open(file_name, "w") as f:
            for i in range(len(integrated)):
                f.write(
                    f"{h[i]:4.0f} {k[i]:4.0f} {l[i]:4.0f} {I[i]:10.2f} {sigI[i]:10.2f}\n"
                )


def write_par_file(file_name, experiment):
    from dxtbx.model import Crystal
    from iotbx.mtz.extract_from_symmetry_lib import ccp4_symbol
    from rstbx.cftbx.coordinate_frame_helpers import align_reference_frame
    from scitbx import matrix

    imageset = experiment.imageset
    detector = imageset.get_detector()
    goniometer = imageset.get_goniometer()
    beam = imageset.get_beam()
    scan = imageset.get_scan()

    R_to_mosflm = align_reference_frame(
        beam.get_s0(), (1.0, 0.0, 0.0), goniometer.get_rotation_axis(), (0.0, 0.0, 1.0)
    )

    cryst = experiment.crystal
    cryst = cryst.change_basis(
        cryst.get_space_group().info().change_of_basis_op_to_reference_setting()
    )
    A = matrix.sqr(cryst.get_A())
    A_inv = A.inverse()

    real_space_a = R_to_mosflm * A_inv.elems[:3]
    real_space_b = R_to_mosflm * A_inv.elems[3:6]
    real_space_c = R_to_mosflm * A_inv.elems[6:9]

    cryst_mosflm = Crystal(
        real_space_a, real_space_b, real_space_c, space_group=cryst.get_space_group()
    )
    U_mosflm = matrix.sqr(cryst_mosflm.get_U())
    B_mosflm = matrix.sqr(cryst_mosflm.get_B())
    UB_mosflm = U_mosflm * B_mosflm
    uc_params = cryst_mosflm.get_unit_cell().parameters()
    assert U_mosflm.is_r3_rotation_matrix(), U_mosflm

    beam_centre = tuple(reversed(detector[0].get_beam_centre(beam.get_s0())))
    distance = detector[0].get_directed_distance()
    polarization = R_to_mosflm * matrix.col(beam.get_polarization_normal())
    rotation = matrix.col(goniometer.get_rotation_axis())
    if rotation.angle(matrix.col(detector[0].get_fast_axis())) < rotation.angle(
        matrix.col(detector[0].get_slow_axis())
    ):
        direction = "FAST"
    else:
        direction = "SLOW"
    rotation = R_to_mosflm * rotation

    # Calculate average spot diameter for SEPARATION parameter
    # http://xds.mpimf-heidelberg.mpg.de/html_doc/xds_parameters.html
    # BEAM_DIVERGENCE=
    # This value is approximately arctan(spot diameter/DETECTOR_DISTANCE)
    profile = experiment.profile
    spot_diameter = math.tan(profile.delta_b() * math.pi / 180) * distance
    spot_diameter_px = spot_diameter * detector[0].get_pixel_size()[0]

    # determine parameters for RASTER keyword
    # http://www.mrc-lmb.cam.ac.uk/harry/cgi-bin/keyword2.cgi?RASTER

    # NXS, NYS (odd integers) define the overall dimensions of the rectangular array of pixels for each spot
    # NXS and NYS are set to twice the spot size plus 5 pixels
    nxs = 2 * int(math.ceil(spot_diameter_px)) + 5
    nys = nxs

    # NRX, NRY are the number of columns or rows of points in the background rim
    # NRX and NRY are set to half the spot size plus 2 pixels
    nrx = int(math.ceil(0.5 * spot_diameter_px)) + 2
    nry = nrx

    # NC the corner background cut-off which corresponds to a half-square of side NC points
    # NC is set to the mean of the spot size in X and Y plus 4
    nc = int(math.ceil(spot_diameter_px)) + 4

    def space_group_symbol(space_group):
        symbol = ccp4_symbol(
            space_group.info(), lib_name="syminfo.lib", require_at_least_one_lib=False
        )
        if symbol != "P 1":
            symbol = symbol.replace(" 1", "")
        symbol = symbol.replace(" ", "")
        return symbol

    logger.info("Saving BEST parameter file to %s", file_name)
    with open(file_name, "w") as f:
        print("# parameter file for BEST", file=f)
        print("TITLE          From DIALS", file=f)
        print("DETECTOR       PILA", file=f)
        print("SITE           Not set", file=f)
        print(
            "DIAMETER       %6.2f"
            % (max(detector[0].get_image_size()) * detector[0].get_pixel_size()[0]),
            file=f,
        )
        print(f"PIXEL          {round(detector[0].get_pixel_size()[0], 10)}", file=f)
        print("ROTAXIS        %4.2f %4.2f %4.2f" % rotation.elems, direction, file=f)
        print("POLAXIS        %4.2f %4.2f %4.2f" % polarization.elems, file=f)
        print("GAIN               1.00", file=f)  # correct for Pilatus images
        # http://strucbio.biologie.uni-konstanz.de/xdswiki/index.php/FAQ#You_said_that_the_XDS_deals_with_high_mosaicity._How_high_mosaicity_is_still_manageable.3F
        # http://journals.iucr.org/d/issues/2012/01/00/wd5161/index.html
        # Transform from XDS definition of sigma_m to FWHM (MOSFLM mosaicity definition)
        print(f"CMOSAIC            {experiment.profile.sigma_m() * 2.355:.2f}", file=f)
        print(f"PHISTART           {scan.get_oscillation_range()[0]:.2f}", file=f)
        print(f"PHIWIDTH           {scan.get_oscillation()[1]:.2f}", file=f)
        print(f"DISTANCE        {distance:7.2f}", file=f)
        print(f"WAVELENGTH      {beam.get_wavelength():.5f}", file=f)
        print(f"POLARISATION    {beam.get_polarization_fraction():7.5f}", file=f)
        print(f"SYMMETRY       {space_group_symbol(cryst.get_space_group())}", file=f)
        print("UB             %9.6f %9.6f %9.6f" % UB_mosflm[:3], file=f)
        print("               %9.6f %9.6f %9.6f" % UB_mosflm[3:6], file=f)
        print("               %9.6f %9.6f %9.6f" % UB_mosflm[6:], file=f)
        print("CELL           %8.2f %8.2f %8.2f %6.2f %6.2f %6.2f" % uc_params, file=f)
        print("RASTER           %i %i %i %i %i" % (nxs, nys, nc, nrx, nry), file=f)
        print(f"SEPARATION      {spot_diameter:.3f}  {spot_diameter:.3f}", file=f)
        print("BEAM           %8.3f %8.3f" % beam_centre, file=f)
        print("# end of parameter file for BEST", file=f)



from __future__ import annotations

import procrunner


def test_run(dials_data, tmp_path):
    procrunner.run(
        (
            "dials.plot_reflections",
            dials_data("centroid_test_data", pathlib=True) / "experiments.json",
            dials_data("centroid_test_data", pathlib=True) / "integrated.refl",
            "scan_range=0,5",
        ),
        working_directory=tmp_path,
    ).check_returncode()
    assert (tmp_path / "centroids.png").is_file()



from __future__ import annotations

from dials_algorithms_filter_ext import *  # noqa: F403; lgtm

__all__ = (  # noqa: F405
    "by_bbox_volume",
    "by_detector_mask",
    "by_resolution_at_centroid",
    "by_shoebox_mask",
    "by_xds_angle",
    "by_xds_small_angle",
    "by_zeta",
    "does_bbox_contain_bad_pixels",
    "is_bbox_outside_image_range",
    "is_bbox_valid",
    "is_xds_angle_valid",
    "is_xds_small_angle_valid",
    "is_zeta_valid",
)



from __future__ import annotations

from scitbx.array_family import flex


def test_complex_double_none_comparison():
    """For unknown reasons this can return a flex.bool() object on some builds."""
    assert (flex.complex_double() == None) is False  # noqa:E711



"""
Reads a libtbx_env file and extract information as a json string.

Some structures e.g. module dictionaries refer to the same object in multiple
places. These will be dumped multiple times. Anything that refers back to a
previous level will show as "__selfreference__". Relocatable paths are just
shown as the regular, joined path (from inspection all base off of the build
path anyway).
"""

from __future__ import annotations

import argparse
import json
import os
import pickle
import sys
from pathlib import Path
from types import ModuleType

BUILD_PATH = None
_all_relocatable_paths = set()
_all_absolute_paths = set()


def _read_obj(obj, prev=None):
    if prev is None:
        prev = []
    if obj in prev:
        return "__selfreference__"
    prev = list(prev) + [obj]

    if isinstance(obj, prop_object):
        dic = {name: _read_obj(val, prev) for name, val in obj.__dict__.items()}
        dic["__type__"] = obj._pathed_type
        return dic
    elif isinstance(obj, list):
        p = []
        for i in obj:
            p.append(_read_obj(i, prev))
        return p
    elif isinstance(obj, dict):
        return {a: _read_obj(b, prev) for a, b in obj.items()}
    else:
        return obj


class prop_object(object):
    """Object that can convert itself to a dictionary"""

    def to_dict(self):
        return _read_obj(self)


def pathed_prop_object(path):
    "Create a class that knows the path it's supposed to represent"

    class _pathed_prop_object(prop_object):
        """Object that can convert itself to a dictionary"""

        _pathed_type = path

    return _pathed_prop_object


class relocatable_path(object):
    def __repr__(self):
        _all_relocatable_paths.add(self)
        return os.path.normpath(os.path.join(str(self._anchor), self.relocatable))


class absolute_path(object):
    def __init__(self, path):
        # This init not used by unpickle - only for rewriting in here
        self._path = str(path)

    def __repr__(self):
        _all_absolute_paths.add(self)
        return self._path


def plainlify(thing):
    if isinstance(thing, (str, int, float, complex)):
        return thing
    if thing in (None, True, False):
        return thing
    if isinstance(thing, tuple):
        return tuple(map(plainlify, thing))
    if isinstance(thing, list):
        return list(map(plainlify, thing))
    if isinstance(thing, dict):
        return {plainlify(key): plainlify(value) for key, value in thing.items()}
    if isinstance(thing, set):
        return [plainlify(item) for item in thing]
    return str(thing)


def new_module(name, doc=None):
    """Create a new module and inject it into sys.modules"""
    m = ModuleType(name, doc)
    m.__file__ = name + ".py"
    sys.modules[name] = m
    return m


# Create the fake libtbx environment
libtbx = new_module("libtbx")
libtbx.env_config = new_module("libtbx.env_config")
libtbx.path = new_module("libtbx.path")
libtbx.env_config.environment = pathed_prop_object("libtbx.env_config.environment")
libtbx.env_config.build_options = pathed_prop_object("libtbx.env_config.build_options")
libtbx.env_config.module = pathed_prop_object("libtbx.env_config.module")
libtbx.path.relocatable_path = relocatable_path
libtbx.path.absolute_path = absolute_path


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Read information from a libtbx_env file"
    )
    parser.add_argument("libtbx_env", type=Path)
    parser.add_argument(
        "--build-path",
        type=Path,
        help=(
            "The actual build path. If this disagrees with the environment, "
            "relative paths will be rewritten. This is to help match the "
            "libtbx behaviour under transplanted installs (e.g. conda)."
        ),
    )
    parser.add_argument(
        "--sys-prefix",
        type=Path,
        help=(
            "The sys.prefix for the distribution the libtbx_env sits within. "
            "If set, and the libtbx_env is an installed one, will rewrite internal prefix."
        ),
    )
    parser.add_argument(
        "--windows",
        action="store_true",
        help=(
            "Apply libtbx-installed-on-windows rewrite-rules, if the libtbx "
            "environment is an installed one. "
            "On windows, libtbx transplants the system root for some paths."
        ),
    )
    args = parser.parse_args()
    if not args.libtbx_env.is_file():
        sys.exit(f"Error: {args.libtbx_env} is not a file")
    # Use the libtbx_env path as the real build_path, for rewriting paths
    BUILD_PATH = args.build_path
    # Load the environment dump
    env = pickle.loads(args.libtbx_env.read_bytes())
    # Because of pickle, we need to __repr__ everything in order to register it
    plainlify(env.to_dict())
    # Rewrite the build folder, if we've been provided one
    if args.build_path:
        orig_build_path = str(env.build_path)
        new_build_path = str(args.build_path.resolve())
        # The relocatable path uses an "absolute path" as a parent. So,
        # make sure we translate all absolute paths that exist, and this
        # means that we've moved all relocatable paths.
        for a_path in _all_absolute_paths:
            if a_path._path == orig_build_path:
                a_path._path = new_build_path

    if env.installed and args.sys_prefix:
        # If this is an installed libtbx_env, then we have rules about rewriting it
        # this is... non-ideal, but since release libtbx_env files are broken on
        # windows, this is the best place to deal with it.
        new_prefix = absolute_path(args.sys_prefix.resolve())
        if args.windows:
            new_prefix = absolute_path(args.sys_prefix.resolve() / "library")
        # Replace the anchor for each of the repository paths
        for path in env.repository_paths:
            path._anchor = new_prefix
        env.bin_path._anchor = new_prefix
        env.exe_path._anchor = new_prefix
        env.include_path._anchor = new_prefix
        env.lib_path._anchor = new_prefix
        env.path_utility._anchor = new_prefix
        # Rewrite all of the module dist paths
        if args.windows:
            for module in env.module_list:
                for dist_path in module.dist_paths:
                    if dist_path is not None:
                        dist_path._anchor = new_prefix

    # Regenerate the plain dictionary now we've rewritten the paths
    d = plainlify(env.to_dict())
    print(json.dumps(d, indent=4))



from __future__ import annotations

import argparse
import copy
import itertools
import logging
import os
import pickle
import sys
import traceback
import warnings
from collections import defaultdict, namedtuple
from glob import glob

from orderedset import OrderedSet

import libtbx.phil
from dxtbx.model import ExperimentList
from dxtbx.model.experiment_list import ExperimentListFactory
from dxtbx.util import get_url_scheme

from dials.array_family import flex
from dials.util import Sorry
from dials.util.multi_dataset_handling import (
    renumber_table_id_columns,
    sort_tables_to_experiments_order,
)
from dials.util.phil import FilenameDataWrapper

logger = logging.getLogger(__name__)

tolerance_phil_scope = libtbx.phil.parse(
    """
tolerance
    .help = "Tolerances used to determine shared models"
    .expert_level = 2
  {

  beam {

    wavelength = 1e-6
      .type = float(value_min=0.0)
      .help = "The wavelength tolerance"

    direction = 1e-6
      .type = float(value_min=0.0)
      .help = "The direction tolerance"

    polarization_normal = 1e-6
      .type = float(value_min=0.0)
      .help = "The polarization normal tolerance"

    polarization_fraction = 1e-6
      .type = float(value_min=0.0)
      .help = "The polarization fraction tolerance"

  }

  detector {

    fast_axis = 1e-6
      .type = float(value_min=0.0)
      .help = "The fast axis tolerance"

    slow_axis = 1e-6
      .type = float(value_min=0.0)
      .help = "The slow axis tolerance"

    origin = 5e-2
      .type = float(value_min=0.0)
      .help = "The origin tolerance"

  }

  goniometer {

    rotation_axis = 1e-6
      .type = float(value_min=0.0)
      .help = "The rotation axis tolerance"

    fixed_rotation = 1e-6
      .type = float(value_min=0.0)
      .help = "The fixed rotation tolerance"

    setting_rotation = 1e-6
      .type = float(value_min=0.0)
      .help = "The setting rotation tolerance"

  }

  scan {

    oscillation = 0.03
      .type = float(value_min=0.0)
      .help = "The oscillation tolerance for the scan, as a fraction of the"
              "image width"

  }
}
"""
)

geometry_phil_scope = libtbx.phil.parse(
    """
geometry
  .help = "Allow overrides of experimental geometry"
  .expert_level = 2
{
  include scope dxtbx.model.beam.beam_phil_scope
  include scope dxtbx.model.detector.detector_phil_scope
  include scope dxtbx.model.goniometer.goniometer_phil_scope
  include scope dxtbx.model.scan.scan_phil_scope

  convert_stills_to_sequences = False
    .type = bool
    .help = "When overriding the scan, convert stills into sequences"
    .short_caption = "Convert stills into sequences"

  convert_sequences_to_stills = False
    .type = bool
    .help = "When overriding the scan, convert sequences into stills"
    .short_caption = "Convert sequences into stills"
}
""",
    process_includes=True,
)


format_phil_scope = libtbx.phil.parse(
    """
format
  .help = "Options to pass to the Format class"
  .expert_level = 2
{
  dynamic_shadowing = auto
    .type = bool
    .help = "Enable dynamic shadowing"
  multi_panel = False
    .type = bool
    .help = "Enable a multi-panel detector model."
            "(Not supported by all detector formats)"
}
"""
)


# Simple tuple to hold basic information on why an argument failed
ArgumentHandlingErrorInfo = namedtuple(
    "ArgumentHandlingErrorInfo",
    ["name", "validation", "message", "traceback", "type", "exception"],
)


class Importer:
    """A class to import the command line arguments."""

    def __init__(
        self,
        args,
        read_experiments=False,
        read_reflections=False,
        read_experiments_from_images=False,
        check_format=True,
        verbose=False,
        compare_beam=None,
        compare_detector=None,
        compare_goniometer=None,
        scan_tolerance=None,
        format_kwargs=None,
        load_models=True,
    ):
        """
        Parse the arguments. Populates its instance attributes in an intelligent way
        from the arguments in args.

        If include is set, only those items set will be tried. If not, then if
        exclude is set, then those items will not be tested.

        These are the types we can import:
         - images: a list of images
         - reflections : a list of reflections
         - experiments: a list of experiments

        :param args: The arguments to parse
        :param read_experiments: Try to read the experiments
        :param read_reflections: Try to read the reflections
        :param read_experiments_from_images: Try to read the experiments from images
        :param check_format: Check the format when reading images
        :param verbose: True/False print out some stuff
        :param load_models: Whether to load all models for ExperimentLists
        """

        # Initialise output
        self.experiments = []
        self.reflections = []
        self.unhandled = args
        # Keep track of any errors whilst handling arguments
        self.handling_errors = defaultdict(list)

        # First try to read image files
        if read_experiments_from_images:
            self.unhandled = self.try_read_experiments_from_images(
                self.unhandled,
                verbose,
                compare_beam,
                compare_detector,
                compare_goniometer,
                scan_tolerance,
                format_kwargs,
                load_models,
            )

        # Second try to read experiment files
        if read_experiments:
            self.unhandled = self.try_read_experiments(
                self.unhandled, check_format, verbose
            )

        # Third try to read reflection files
        if read_reflections:
            self.unhandled = self.try_read_reflections(self.unhandled, verbose)

    def _handle_converter_error(self, argument, exception, type, validation=False):
        "Record information about errors that occurred processing an argument"
        self.handling_errors[argument].append(
            ArgumentHandlingErrorInfo(
                name=argument,
                validation=validation,
                message=str(exception),
                traceback=traceback.format_exc(),
                type=type,
                exception=exception,
            )
        )

    def try_read_experiments_from_images(
        self,
        args,
        verbose,
        compare_beam,
        compare_detector,
        compare_goniometer,
        scan_tolerance,
        format_kwargs,
        load_models=True,
    ):
        """
        Try to import images.

        :param args: The input arguments
        :param verbose: Print verbose output
        :param compare_beam:
        :param compare_detector:
        :param compare_goniometer:
        :param scan_tolerance:
        :param format_kwargs:
        :param load_models: Whether to load all models for ExperimentLists
        :return: Unhandled arguments
        """
        from dxtbx.model.experiment_list import ExperimentListFactory

        # If filenames contain wildcards, expand
        args_new = []
        for arg in args:
            # Don't expand wildcards if URI-style filename
            if "*" in arg and not get_url_scheme(arg):
                filenames = glob(arg)
                if filenames:
                    args_new.extend(filenames)
                else:
                    logger.warning(f"No files found with pattern {arg}")
            else:
                args_new.append(arg)
        args = args_new

        unhandled = []

        try:
            experiments = ExperimentListFactory.from_filenames(
                args,
                unhandled=unhandled,
                compare_beam=compare_beam,
                compare_detector=compare_detector,
                compare_goniometer=compare_goniometer,
                scan_tolerance=scan_tolerance,
                format_kwargs=format_kwargs,
                load_models=load_models,
            )
        except FileNotFoundError as e:
            logger.error(f"File {e.filename} not found")
        else:
            if experiments:
                self.experiments.append(
                    FilenameDataWrapper(filename="<image files>", data=experiments)
                )

        return unhandled

    def try_read_experiments(self, args, check_format, verbose):
        """
        Try to import experiments.

        :param args: The input arguments
        :param check_format: True/False check the image format
        :param verbose: Print verbose output
        :returns: Unhandled arguments
        """
        from dxtbx.model.experiment_list import InvalidExperimentListError

        unhandled = []
        for argument in args:
            try:
                self.experiments.append(
                    FilenameDataWrapper(
                        filename=argument,
                        data=ExperimentListFactory.from_json_file(
                            argument, check_format=check_format
                        ),
                    )
                )
            except InvalidExperimentListError as e:
                # This is a validation-related error: The file appears not to be in the correct format
                self._handle_converter_error(
                    argument, e, type="ExperimentList", validation=True
                )
                unhandled.append(argument)
            except Exception as e:
                self._handle_converter_error(argument, e, type="ExperimentList")
                unhandled.append(argument)
        return unhandled

    def try_read_reflections(self, args, verbose):
        """Try to import reflections.

        :param args: The input arguments
        :param verbose: Print verbose output
        :returns: Unhandled arguments
        """
        unhandled = []
        for argument in args:
            try:
                if not os.path.exists(argument):
                    raise Sorry(f"File {argument} does not exist")
                self.reflections.append(
                    FilenameDataWrapper(
                        filename=argument,
                        data=flex.reflection_table.from_file(argument),
                    )
                )
            except pickle.UnpicklingError:
                self._handle_converter_error(
                    argument,
                    pickle.UnpicklingError("Appears to be an invalid pickle file"),
                    type="Reflections",
                    validation=True,
                )
                unhandled.append(argument)
            except Exception as e:
                self._handle_converter_error(argument, e, type="Reflections")
                unhandled.append(argument)
        return unhandled


class PhilCommandParser:
    """A class to parse phil parameters from positional arguments"""

    def __init__(
        self,
        phil=None,
        read_experiments=False,
        read_reflections=False,
        read_experiments_from_images=False,
        check_format=True,
    ):
        """
        Initialise the parser.

        :param phil: The phil scope
        :param read_experiments: Try to read the experiments
        :param read_reflections: Try to read the reflections
        :param read_experiments_from_images: Try to read the experiments from images
        :param check_format: Check the format when reading images
        """
        from dials.util.phil import parse

        # Set the system phil scope
        if phil is None:
            self._system_phil = parse("")
        else:
            self._system_phil = copy.deepcopy(phil)

        # Set the flags
        self._read_experiments = read_experiments
        self._read_reflections = read_reflections
        self._read_experiments_from_images = read_experiments_from_images
        self._check_format = check_format

        # Adopt the input scope
        input_phil_scope = self._generate_input_scope()
        if input_phil_scope is not None:
            self.system_phil.adopt_scope(input_phil_scope)

        # Set the working phil scope
        self._phil = self.system_phil.fetch(source=parse(""))

    @property
    def phil(self):
        """
        Get the phil object

        :return: The phil scope
        """
        return self._phil

    @property
    def system_phil(self):
        """
        Get the system phil.

        :return: The system phil scope
        """
        return self._system_phil

    @property
    def diff_phil(self):
        """
        Get the diff phil.

        :return: The difference phil scope
        """
        return self.system_phil.fetch_diff(source=self.phil)

    def parse_args(
        self, args, verbose=False, return_unhandled=False, quick_parse=False
    ):
        """
        Parse the command line arguments.

        :param args: The input arguments
        :param verbose: Print verbose output
        :param return_unhandled: True/False also return unhandled arguments
        :param quick_parse: Return as fast as possible and without reading any data,
                            ignoring class constructor options.
        :return: The options and parameters and (optionally) unhandled arguments
        """
        from dxtbx.model.experiment_list import (
            BeamComparison,
            DetectorComparison,
            GoniometerComparison,
        )

        from dials.util.phil import parse

        # Parse the command line phil parameters
        user_phils = []
        unhandled = []
        interpretor = self.system_phil.command_line_argument_interpreter()

        def _is_a_phil_file(filename):
            return any(
                filename.endswith(phil_ext)
                for phil_ext in (".phil", ".param", ".params", ".eff", ".def")
            )

        for arg in args:
            if (
                _is_a_phil_file(arg)
                and os.path.isfile(arg)
                and os.path.getsize(arg) > 0
            ):
                try:
                    user_phils.append(parse(file_name=arg))
                except Exception:
                    if return_unhandled:
                        unhandled.append(arg)
                    else:
                        raise
            # Treat "has a schema" as "looks like a URL (not phil)
            elif "=" in arg and not get_url_scheme(arg):
                try:
                    user_phils.append(interpretor.process_arg(arg=arg))
                except Exception:
                    if return_unhandled:
                        unhandled.append(arg)
                    else:
                        raise
            else:
                unhandled.append(arg)

        # Fetch the phil parameters
        self._phil, unused = self.system_phil.fetch(
            sources=user_phils, track_unused_definitions=True
        )

        # Print if bad definitions
        if len(unused) > 0:
            msg = [item.object.as_str().strip() for item in unused]
            msg = "\n".join(["  %s" % line for line in msg])
            raise RuntimeError(f"The following definitions were not recognised\n{msg}")

        # Extract the parameters
        params = self._phil.extract()

        # Stop at this point if quick_parse is set. A second pass may be needed.
        if quick_parse:
            return params, unhandled

        # Create some comparison functions
        if self._read_experiments_from_images:
            compare_beam = BeamComparison(
                wavelength_tolerance=params.input.tolerance.beam.wavelength,
                direction_tolerance=params.input.tolerance.beam.direction,
                polarization_normal_tolerance=params.input.tolerance.beam.polarization_normal,
                polarization_fraction_tolerance=params.input.tolerance.beam.polarization_fraction,
            )
            compare_detector = DetectorComparison(
                fast_axis_tolerance=params.input.tolerance.detector.fast_axis,
                slow_axis_tolerance=params.input.tolerance.detector.slow_axis,
                origin_tolerance=params.input.tolerance.detector.origin,
            )
            compare_goniometer = GoniometerComparison(
                rotation_axis_tolerance=params.input.tolerance.goniometer.rotation_axis,
                fixed_rotation_tolerance=params.input.tolerance.goniometer.fixed_rotation,
                setting_rotation_tolerance=params.input.tolerance.goniometer.setting_rotation,
            )
            scan_tolerance = params.input.tolerance.scan.oscillation

            # FIXME Should probably make this smarter since it requires editing here
            # and in dials.import phil scope
            try:
                format_kwargs = {
                    "dynamic_shadowing": params.format.dynamic_shadowing,
                    "multi_panel": params.format.multi_panel,
                }
            except AttributeError:
                format_kwargs = None
        else:
            compare_beam = None
            compare_detector = None
            compare_goniometer = None
            scan_tolerance = None
            format_kwargs = None

        try:
            load_models = params.load_models
        except AttributeError:
            load_models = True

        # Try to import everything
        importer = Importer(
            unhandled,
            read_experiments=self._read_experiments,
            read_reflections=self._read_reflections,
            read_experiments_from_images=self._read_experiments_from_images,
            check_format=self._check_format,
            verbose=verbose,
            compare_beam=compare_beam,
            compare_detector=compare_detector,
            compare_goniometer=compare_goniometer,
            scan_tolerance=scan_tolerance,
            format_kwargs=format_kwargs,
            load_models=load_models,
        )

        # Grab a copy of the errors that occurred in case the caller wants them
        self.handling_errors = importer.handling_errors

        # Add the cached arguments
        for obj in importer.experiments:
            params.input.experiments.append(obj)
        for obj in importer.reflections:
            params.input.reflections.append(obj)

        # Convert to phil
        self._phil = self.system_phil.format(python_object=params)

        return params, importer.unhandled

    def _generate_input_scope(self):
        """
        Generate the required input scope.

        :return: The input phil scope
        """
        from dials.util.phil import parse

        # Create the input scope
        require_input_scope = (
            self._read_experiments
            or self._read_reflections
            or self._read_experiments_from_images
        )
        if not require_input_scope:
            return None
        input_phil_scope = parse("input {}")
        main_scope = input_phil_scope.get_without_substitution("input")
        assert len(main_scope) == 1
        main_scope = main_scope[0]

        # Add the experiments phil scope
        if self._read_experiments or self._read_experiments_from_images:
            phil_scope = parse(
                f"""
        experiments = None
          .type = experiment_list(check_format={self._check_format!r})
          .multiple = True
          .help = "The experiment list file path"
      """
            )
            main_scope.adopt_scope(phil_scope)

        # If reading images, add some more parameters
        if self._read_experiments_from_images:
            main_scope.adopt_scope(tolerance_phil_scope)

        # Add the reflections scope
        if self._read_reflections:
            phil_scope = parse(
                """
        reflections = None
          .type = reflection_table
          .multiple = True
          .help = "The reflection table file path"
      """
            )
            main_scope.adopt_scope(phil_scope)

        # Return the input scope
        return input_phil_scope


class ArgumentParserBase(argparse.ArgumentParser):
    """The base class for the option parser."""

    def __init__(self, config_options=False, sort_options=False, **kwargs):
        """
        Initialise the class.

        :param config_options: True/False show configuration options
        :param sort_options: True/False show argument sorting options
        """

        # Initialise the option parser
        super().__init__(add_help=False, **kwargs)

        # Add an option to show configuration parameters
        if config_options:
            self.add_argument(
                "-c",
                "--show-config",
                action="store_true",
                default=False,
                dest="show_config",
                help="Show the configuration parameters.",
            )
            self.add_argument(
                "-a",
                "--attributes-level",
                default=0,
                type=int,
                dest="attributes_level",
                help="Set the attributes level for showing configuration parameters",
            )
            self.add_argument(
                "-e",
                "--expert-level",
                type=int,
                default=0,
                dest="expert_level",
                help="Set the expert level for showing configuration parameters",
            )
            self.add_argument(
                "--export-autocomplete-hints",
                action="store_true",
                default=False,
                dest="export_autocomplete_hints",
                help=argparse.SUPPRESS,
            )

        # Add an option to sort
        if sort_options:
            self.add_argument(
                "-s",
                "--sort",
                action="store_true",
                dest="sort",
                default=False,
                help="Sort the arguments",
            )

        # Set a help parameter
        self.add_argument(
            "-h",
            "--help",
            action="count",
            default=0,
            dest="help",
            help="Show this help message and exit. Can be specified multiple times to increase verbosity.",
        )

        # Set a verbosity parameter
        self.add_argument(
            "-v",
            action="count",
            default=0,
            dest="verbose",
            help="Increase verbosity (can be specified multiple times)",
        )

        # Add an option for PHIL file to parse - PHIL files passed as
        # positional arguments are also read but this allows the user to
        # explicitly specify STDIN
        self.add_argument(
            "--phil",
            action="append",
            metavar="FILE",
            help="PHIL files to read. Pass '-' for STDIN. Can be specified multiple times, but duplicates ignored.",
        )

    def parse_known_args(self, args=None, quick_parse=False):
        """
        Parse the command line arguments and get system configuration.

        :param args: The arguments to parse.
        :returns: The options and phil parameters
        """

        # Parse the command line arguments, this will separate out
        # options (e.g. -o, --option) and positional arguments, in
        # which phil options will be included.
        options, args = super().parse_known_args(args=args)

        # Read any argument-specified PHIL file. Ignore duplicates.
        if options.phil:
            for philfile in OrderedSet(options.phil):
                # Should we read STDIN?
                if philfile == "-":
                    lines = sys.stdin.readlines()
                else:
                    # Otherwise, assume we've been given a path
                    with open(philfile) as phil_input:
                        lines = phil_input.readlines()
                # Add these to the unparsed argument list
                args.extend(l.strip() for l in lines)

        # Maybe sort the data
        if hasattr(options, "sort") and options.sort:
            args = sorted(args)

        # Return the parameters
        return options, args

    def format_epilog(self, formatter):
        """Don't do formatting on epilog."""
        if self.epilog is None:
            return ""
        return self.epilog


class ArgumentParser(ArgumentParserBase):
    """A class to parse command line options and get the system configuration.
    The class extends argparse.ArgumentParser to include the reading of phil
    parameters."""

    def __init__(
        self,
        phil=None,
        read_experiments=False,
        read_reflections=False,
        read_experiments_from_images=False,
        check_format=True,
        sort_options=False,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        **kwargs,
    ):
        """
        Initialise the class.

        :param phil: The phil scope
        :param read_experiments: Try to read the experiments
        :param read_reflections: Try to read the reflections
        :param read_experiments_from_images: Try to read the experiments from images
        :param check_format: Check the format when reading images
        :param sort_options: Show argument sorting options
        """

        # Create the phil parser
        self._phil_parser = PhilCommandParser(
            phil=phil,
            read_experiments=read_experiments,
            read_reflections=read_reflections,
            read_experiments_from_images=read_experiments_from_images,
            check_format=check_format,
        )

        # Initialise the option parser
        super().__init__(
            sort_options=sort_options,
            config_options=self.system_phil.as_str() != "",
            formatter_class=formatter_class,
            **kwargs,
        )

    def parse_args(
        self,
        args=None,
        show_diff_phil=False,
        return_unhandled=False,
        ignore_unhandled=False,
        quick_parse=False,
    ):
        """
        Parse the command line arguments and get system configuration.

        :param args: The input arguments
        :param show_diff_phil: True/False Print the diff phil
        :param return_unhandled: True/False return unhandled arguments
        :param ignore_unhandled: True/False ignore unhandled arguments
                                  if return_unhandled is False
        :param quick_parse: Return as fast as possible and without reading any data,
                            ignoring class constructor options
        :return: The options and phil parameters
        """

        # Parse the command line arguments, this will separate out
        # options (e.g. -o, --option) and positional arguments, in
        # which phil options will be included.
        options, args = super().parse_known_args(args=args, quick_parse=quick_parse)

        if options.help:
            self.print_help()

        # Show config
        if hasattr(options, "show_config") and options.show_config:
            show_config = True
            attributes_level = options.attributes_level
            expert_level = options.expert_level
        elif options.help:
            show_config = True
            attributes_level = options.verbose
            expert_level = options.help - 1
        else:
            show_config = False

        if show_config:
            print(
                "Showing configuration parameters with:\n"
                f"  attributes_level = {attributes_level}\n"
                f"  expert_level = {expert_level}\n"
            )
            print(
                self.phil.as_str(
                    expert_level=expert_level,
                    attributes_level=attributes_level,
                )
            )
            exit(0)

        if (
            hasattr(options, "export_autocomplete_hints")
            and options.export_autocomplete_hints
        ):
            self._export_autocomplete_hints()
            exit(0)

        # Parse the phil parameters
        params, args = self._phil_parser.parse_args(
            args,
            options.verbose > 0,
            return_unhandled=return_unhandled,
            quick_parse=quick_parse,
        )

        # Print the diff phil
        if show_diff_phil:
            diff_phil_str = self.diff_phil.as_str()
            if diff_phil_str != "":
                print("The following parameters have been modified:\n")
                print(diff_phil_str)

        # Return the parameters
        if return_unhandled:
            return params, options, args
        elif len(args) > 0 and not quick_parse:
            # Handle printing any messages to diagnose unhandled arguments
            msg = self._warn_about_unhandled_args(args, verbosity=options.verbose)

            if ignore_unhandled:
                print(msg)
            else:
                raise Sorry(msg)
        return params, options

    def _warn_about_unhandled_args(self, unhandled, verbosity=0):
        """
        Generate any messages about unhandled arguments.

        This separates errors by validation/non-validation related, and only
        gives the user validation information if there's no other reason (or
        asked for verbose output).

        :param unhandled: List of unhandled arguments
        :param verbosity: The output verbosity determined during parsing
        :returns:         A formatted information string
        """
        msg = []
        msg.append("Unable to handle the following arguments:")

        # If we have any detailed information about why any of the
        # arguments weren't processed, give this to the user
        for arg in [x for x in unhandled if x in self._phil_parser.handling_errors]:
            # Split the reasons for unhandling into validation, non-validation
            non_valid = [
                x for x in self._phil_parser.handling_errors[arg] if not x.validation
            ]
            valid = [x for x in self._phil_parser.handling_errors[arg] if x.validation]

            # If we have non-validation-related errors, these are more important
            for _, err in itertools.groupby(non_valid, key=lambda x: x.message):
                err = list(err)
                # Grouping the errors by message lets us avoid repeating messages
                if len(err) > 1:
                    msg.append(
                        '  "{}" failed repeatedly during processing:\n{}\n'.format(
                            arg, "    " + err[0].message
                        )
                    )
                elif isinstance(err[0].exception, Sorry):
                    msg.append(
                        '  "{}" failed during {} processing:\n    {}\n'.format(
                            arg, err[0].type, err[0].message
                        )
                    )
                else:
                    msg.append(
                        '  "{}" failed during {} processing:\n{}\n'.format(
                            arg,
                            err[0].type,
                            "\n".join(
                                "    " + x for x in err[0].traceback.splitlines()
                            ),
                        )
                    )

            # Otherwise (or if asked for verbosity), list the validation errors
            if valid and (not non_valid or verbosity > 1):
                msg.append(
                    "  {} did not appear to conform to any{} expected format:".format(
                        arg, " other" if non_valid else ""
                    )
                )
                slen = max(len(x.type) for x in valid)
                for err in valid:
                    msg.append(f"    - {f'{err.type}:'.ljust(slen + 1)} {err.message}")
        # The others
        for arg in [x for x in unhandled if x not in self._phil_parser.handling_errors]:
            msg.append("  " + str(arg))
        return "\n".join(msg)

    @property
    def phil(self):
        """
        Get the phil object

        :returns: The phil scope
        """
        return self._phil_parser.phil

    @property
    def system_phil(self):
        """
        Get the system phil.

        :returns: The system phil scope
        """
        return self._phil_parser.system_phil

    @property
    def diff_phil(self):
        """
        Get the diff phil.

        :returns: The diff phil scope
        """
        return self._phil_parser.diff_phil

    def _strip_rst_markup(self, text):
        """
        Strip rst markup

        :param text: The text to strip
        :return: The stripped text
        """
        return text.replace("::", ":")

    def format_help(self):
        """
        Format the help string

        :param formatter: The formatter to use
        :return: The formatted help text
        """
        result = super().format_help()
        return self._strip_rst_markup(result)

    def _export_autocomplete_hints(self):
        # complete list of all parameters
        parameter_list = []
        # short name -> full name expansion for unique names
        parameter_expansion_list = {}
        # full name -> list of flags for choice parameters
        parameter_choice_list = {}

        for d in self.phil.all_definitions():
            # Create complete list of all parameters
            parameter_list.append(d.path)

            # Create expansion (alias) list for unique names that are not expert commands
            if (
                d.object.name not in parameter_expansion_list
                and d.object.expert_level is None
                and d.parent.expert_level is None
            ):
                parameter_expansion_list[d.object.name] = d.path
            else:
                parameter_expansion_list[d.object.name] = None

            # Extract parameter choice lists
            if d.object.type.phil_type == "choice":
                parameter_choice_list[d.path] = [
                    w[1:] if w.startswith("*") else w
                    for w in (str(x) for x in d.object.words)
                ]
            elif d.object.type.phil_type == "bool":
                parameter_choice_list[d.path] = ["true", "false"]

        def construct_completion_tree(paths):
            """Construct a tree of parameters, grouped by common prefixes"""

            # Split parameter paths at '.' character
            paths = [p.split(".", 1) for p in paths]

            # Identify all names that are directly on this level
            # or represent parameter groups with a common prefix
            top_elements = {f"{x[0]}{'=' if len(x) == 1 else '.'}" for x in paths}

            # Partition all names that are further down the tree by their prefix
            subpaths = {}
            for p in paths:
                if len(p) > 1:
                    if p[0] not in subpaths:
                        subpaths[p[0]] = []
                    subpaths[p[0]].append(p[1])

            # If there are prefixes with only one name beneath them, put them on the top level
            for s in list(subpaths.keys()):
                if len(subpaths[s]) == 1:
                    top_elements.remove(f"{s}.")
                    top_elements.add(f"{s}.{subpaths[s][0]}=")
                    del subpaths[s]

            result = {"": top_elements}
            # Revursively process each group
            for n, x in subpaths.items():
                result[n] = construct_completion_tree(x)

            return result

        print("function _dials_autocomplete_flags ()")
        print("{")
        print(' case "$1" in')
        for p in parameter_choice_list:
            print(f"\n  {p})")
            print(
                '   _dials_autocomplete_values="%s";;'
                % " ".join(parameter_choice_list[p])
            )
        print("\n  *)")
        print('    _dials_autocomplete_values="";;')
        print(" esac")
        print("}")

        print("function _dials_autocomplete_expansion ()")
        print("{")
        print(' case "$1" in')
        for p, exp in parameter_expansion_list.items():
            if exp is not None:
                print(f"\n  {p}=)")
                print(f'   _dials_autocomplete_values="{exp}=";;')
        print("\n  *)")
        print('    _dials_autocomplete_values="";;')
        print(" esac")
        print("}")

        tree = construct_completion_tree(parameter_list)

        def _tree_to_bash(prefix, tree):
            for subkey in tree:
                if subkey != "":
                    _tree_to_bash(prefix + subkey + ".", tree[subkey])
                    print(f"\n  {prefix + subkey + '.'}*)")
                    print(
                        '    _dials_autocomplete_values="%s";;'
                        % " ".join(
                            sorted(
                                [prefix + subkey + "." + x for x in tree[subkey][""]]
                            )
                        )
                    )

        print("function _dials_autocomplete_hints ()")
        print("{")
        print(' case "$1" in')
        _tree_to_bash("", tree)

        toplevelset = tree[""] | {
            p + "=" for p, exp in parameter_expansion_list.items() if exp is not None
        }

        print("\n  *)")
        print('    _dials_autocomplete_values="%s";;' % " ".join(sorted(toplevelset)))
        print(" esac")
        print("}")


class OptionParser(ArgumentParser):
    def __init__(self, *args, **kwargs):
        # Backwards compatibility 2021-11-10; 2022-04-06
        # Remove after Dec 2022
        warnings.warn(
            "OptionParser is deprecated, use ArgumentParser instead",
            UserWarning,
            stacklevel=2,
        )
        super().__init__(*args, **kwargs)


def flatten_reflections(filename_object_list):
    """
    Flatten a list of reflections tables

    A check is also made for the 'id' values in the reflection tables, which are
    renumbered from 0..n-1 to avoid clashes. The experiment_identifiers dict is
    also updated if present in the input tables.

    :param filename_object_list: The parameter item
    :return: The flattened reflection table
    """
    tables = [o.data for o in filename_object_list]
    if len(tables) > 1:
        tables = renumber_table_id_columns(tables)
    return tables


def flatten_experiments(filename_object_list):
    """
    Flatten a list of experiment lists

    :param filename_object_list: The parameter item
    :return: The flattened experiment lists
    """

    result = ExperimentList()
    for o in filename_object_list:
        result.extend(o.data)
    return result


def reflections_and_experiments_from_files(
    reflection_file_object_list, experiment_file_object_list
):
    """Extract reflection tables and an experiment list from the files.
    If experiment identifiers are set, the order of the reflection tables is
    changed to match the order of experiments.
    """
    tables = flatten_reflections(reflection_file_object_list)

    experiments = flatten_experiments(experiment_file_object_list)

    if tables and experiments:
        tables = sort_tables_to_experiments_order(tables, experiments)

    return tables, experiments



from __future__ import annotations


class GaussianRSMaskCalculatorFactory:
    """
    Factory class for mask calculator
    """

    @staticmethod
    def create(experiments):
        """
        Create the mask calculator
        """
        from dials.algorithms.integration.parallel_integrator import (
            GaussianRSMaskCalculator,
            GaussianRSMultiCrystalMaskCalculator,
        )

        result = GaussianRSMultiCrystalMaskCalculator()
        for e in experiments:
            alg = GaussianRSMaskCalculator(
                e.beam,
                e.detector,
                e.goniometer,
                e.scan,
                e.profile.delta_b(deg=False),
                e.profile.delta_m(deg=False),
            )
            result.append(alg)
        return result


class GaussianRSIntensityCalculatorFactory:
    """
    A class to create the intensity calculator
    """

    @staticmethod
    def create(data, detector_space=False, deconvolution=False):
        """
        Create the intensity calculator
        """
        from dials.algorithms.integration.parallel_integrator import (
            GaussianRSIntensityCalculator,
        )

        # Return the intensity algorithm
        return GaussianRSIntensityCalculator(data, detector_space, deconvolution)


class GaussianRSReferenceCalculatorFactory:
    """
    A class to create the reference calculator
    """

    @staticmethod
    def create(experiments, grid_size=5, scan_step=5, grid_method="circular_grid"):
        """
        Create the intensity calculator
        """
        from math import ceil

        from dials.algorithms.integration.parallel_integrator import (
            GaussianRSReferenceCalculator,
        )
        from dials.algorithms.profile_model.gaussian_rs.transform import TransformSpec
        from dials.algorithms.profile_model.modeller import (
            CircleSampler,
            GridSampler,
            SingleSampler,
        )

        # Assume the detector and scan are the same in each case
        detector = experiments[0].detector
        scan = experiments[0].scan

        # Get the number of scan points
        scan_range = scan.get_oscillation_range(deg=True)
        scan_range = abs(scan_range[1] - scan_range[0])
        num_scan_points = int(ceil(scan_range / scan_step))

        # If multi panel then set to single
        if grid_method in ["regular_grid", "circular_grid"] and len(detector) > 1:
            grid_method = "single"

        # Create the sampler
        if grid_method == "single":
            sampler = SingleSampler(scan.get_array_range(), num_scan_points)
        elif grid_method == "regular_grid":
            sampler = GridSampler(
                detector[0].get_image_size(),
                scan.get_array_range(),
                (3, 3, num_scan_points),
            )
        elif grid_method == "circular_grid":
            sampler = CircleSampler(
                detector[0].get_image_size(), scan.get_array_range(), num_scan_points
            )
        else:
            raise RuntimeError("Unknown grid type")

        # Create the spec list
        spec_list = []
        for experiment in experiments:

            spec = TransformSpec(
                experiment.beam,
                experiment.detector,
                experiment.goniometer,
                experiment.scan,
                experiment.profile.sigma_b(deg=False),
                experiment.profile.sigma_m(deg=False),
                experiment.profile.n_sigma() * 1.5,
                grid_size,
            )

            spec_list.append(spec)

        # Return the intensity algorithm
        return GaussianRSReferenceCalculator(sampler, spec_list)



from __future__ import annotations

import logging

from jinja2 import ChoiceLoader, Environment, PackageLoader

from cctbx import uctbx

from dials.algorithms.clustering import plots as cluster_plotter
from dials.algorithms.clustering.observers import uc_params_from_experiments
from dials.algorithms.merging.merge import MergingStatisticsData
from dials.algorithms.scaling.observers import make_merging_stats_plots
from dials.array_family import flex
from dials.report.plots import d_star_sq_to_d_ticks
from dials.util import tabulate

logger = logging.getLogger("dials")


class MergeJSONCollector(object):

    initiated = False
    data = {}

    @classmethod
    def __init__(cls):
        cls.initiated = True

    @classmethod
    def reset(cls):
        cls.data = {}
        cls.initiated = False

    def create_json(self):
        return generate_json_data(self.data)


def generate_json_data(data: dict[float, MergingStatisticsData]) -> dict:
    json_data = {}
    for wl, stats in data.items():
        wl_key = f"{wl:.5f}"
        stats_plots = make_merging_stats_plots(
            stats,
            run_xtriage_analysis=True,
            make_batch_plots=False,
        )
        # remove unneeded items
        for i in ["batch_plots", "anom_plots", "image_range_tables"]:
            del stats_plots[i]
        uc_params = uc_params_from_experiments(stats.experiments)
        stats_plots["unit_cell_plots"] = cluster_plotter.plot_uc_histograms(
            uc_params, scatter_style="heatmap"
        )
        json_data[wl_key] = stats_plots
        if stats.anomalous_amplitudes:
            json_data[wl_key]["resolution_plots"].update(
                make_dano_plots({wl: stats.anomalous_amplitudes})["dF"]
            )
        json_data[wl_key]["merging_stats"] = stats.merging_statistics_result.as_dict()
        json_data[wl_key][
            "merging_stats_anom"
        ] = stats.anom_merging_statistics_result.as_dict()
    if len(json_data) > 1:
        # create an overall summary table
        headers = [""] + ["Wavelength " + f"{wl:.5f}" + " Å" for wl in data.keys()]
        rows = []
        first_wl = f"{list(data.keys())[0]:.5f}"
        for k, v in json_data[first_wl]["scaling_tables"][
            "overall_summary_data"
        ].items():
            rows.append([k, v])
        for wl in list(data.keys())[1:]:
            for i, v in enumerate(
                json_data[f"{wl:.5f}"]["scaling_tables"][
                    "overall_summary_data"
                ].values()
            ):
                rows[i].append(v)
        json_data["multi_data"] = {"overall_summary": [headers] + rows}
    return json_data


def dano_over_sigdano_stats(anomalous_amplitudes, n_bins=20):
    """Calculate the statistic for resolution bins and overall."""
    vals = flex.double()
    # First calculate dF/s(dF) per resolution bin
    try:
        anomalous_amplitudes.setup_binner_counting_sorted(n_bins=n_bins)
    except Exception as e:
        logger.info(f"Unable to setup binner successfully, error: {e}")
        return None, None
    else:
        resolution_bin_edges = flex.double()
        for i_bin in anomalous_amplitudes.binner().range_used():
            sel = anomalous_amplitudes.binner().selection(i_bin)
            arr = anomalous_amplitudes.select(sel)
            vals.append(dano_over_sigdano(arr))
            resolution_bin_edges.append(anomalous_amplitudes.binner().bin_d_min(i_bin))
        resolution_bin_edges.append(anomalous_amplitudes.binner().d_min())
        return vals, resolution_bin_edges


def dano_over_sigdano(anomalous_amplitudes):
    """Calculate < |F(+) - F(-)| / sigma(F(+) - F(-))> i.e. <DANO/SIGDANO>."""
    diff = anomalous_amplitudes.anomalous_differences()
    if not diff.data() or not diff.sigmas():
        return 0.0
    return flex.mean(flex.abs(diff.data()) / diff.sigmas())


def make_dano_table(anomalous_amplitudes):
    """Calculate <dano/sigdano> in resolution bins and tabulate."""
    dFsdF, resolution_bin_edges = dano_over_sigdano_stats(anomalous_amplitudes)
    if not dFsdF:
        return ""
    logger.info("Size of anomalous differences")
    header = ["d_max", "d_min", "<|ΔF|/σ(ΔF)>"]
    rows = []
    for i, dF in enumerate(dFsdF):
        rows.append(
            [
                f"{resolution_bin_edges[i]:6.2f}",
                f"{resolution_bin_edges[i+1]:6.2f}",
                f"{dF:6.3f}",
            ]
        )
    return tabulate(rows, header)


def make_dano_plots(anomalous_data):
    """
    Make dicts of data for plotting e.g. for plotly.

    Args:
        anomalous_data (dict) : A dict of (wavelength, anomalous array) data.

    Returns:
        dict: A dictionary containing the plotting data.
    """

    data = {
        "dF": {
            "dano": {
                "data": [],
                "help": """\
This plot shows the size of the anomalous differences of F relative to the uncertainties,
(<|F(+)-F(-)|/σ(F(+)-F(-))>). A value of 0.8 is indicative of pure noise, and
a suggested cutoff is when the value falls below 1.2, although these guides require
reliable sigma estimates. For further information see
https://strucbio.biologie.uni-konstanz.de/ccp4wiki/index.php?title=SHELX_C/D/E
""",
                "layout": {"title": "<|ΔF|/σ(ΔF)> vs resolution"},
            },
        },
    }

    for i, (wave, anom) in enumerate(anomalous_data.items()):
        dFsdF, resolution_bin_edges = dano_over_sigdano_stats(anom)
        if not dFsdF:
            continue
        d_star_sq_bins = [
            0.5
            * (
                uctbx.d_as_d_star_sq(resolution_bin_edges[i])
                + uctbx.d_as_d_star_sq(resolution_bin_edges[i + 1])
            )
            for i in range(0, len(resolution_bin_edges[:-1]))
        ]
        d_star_sq_tickvals, d_star_sq_ticktext = d_star_sq_to_d_ticks(
            d_star_sq_bins, nticks=5
        )
        data["dF"]["dano"]["data"].append(
            {
                "x": d_star_sq_bins,
                "y": list(dFsdF),
                "type": "scatter",
                "name": "\u03BB" + f"={wave:.4f}",
            }
        )
    if not data["dF"]["dano"]["data"]:
        return data
    data["dF"]["dano"]["data"].append(
        {
            "x": d_star_sq_bins,
            "y": [0.8] * len(d_star_sq_bins),
            "type": "scatter",
            "mode": "lines",
            "name": "random noise level",
        }
    )
    data["dF"]["dano"]["data"].append(
        {
            "x": d_star_sq_bins,
            "y": [1.2] * len(d_star_sq_bins),
            "type": "scatter",
            "mode": "lines",
            "name": "an approximate <br>threshold for a<br>resolution cutoff",
        }
    )

    data["dF"]["dano"]["layout"] = {
        "title": "<|ΔF|/σ(ΔF)> vs resolution",
        "xaxis": {
            "title": "Resolution (Å)",
            "tickvals": d_star_sq_tickvals,
            "ticktext": d_star_sq_ticktext,
        },
        "yaxis": {"title": "<|ΔF|/σ(ΔF)>", "rangemode": "tozero"},
    }
    return data


def generate_html_report(json_data, filename):
    multi_data = None
    if "multi_data" in json_data:
        multi_data = json_data.pop("multi_data")
    for wl, v in json_data.items():
        str_wl = wl.replace(".", "_")
        for plot_cat in ["resolution_plots", "misc_plots"]:
            for name in list(v[plot_cat].keys()):
                v[plot_cat][name + "_" + str_wl] = v[plot_cat].pop(name)

    loader = ChoiceLoader(
        [
            PackageLoader("dials", "templates"),
            PackageLoader("dials", "static", encoding="utf-8"),
        ]
    )
    env = Environment(loader=loader)
    template = env.get_template("merge_report.html")
    html = template.render(
        page_title="DIALS merge report",
        individual_reports=json_data,
        multi_data=multi_data,
    )
    logger.info("Writing html report to %s", filename)
    with open(filename, "wb") as f:
        f.write(html.encode("utf-8", "xmlcharrefreplace"))
